import abc
import itertools
from enum import Enum
from typing import Callable, Iterable, List, Optional, Tuple

from github3.exceptions import NotFoundError

from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.dependencies.dependencies import (
    BaseGitHubDependency,
    Dependency,
    DynamicDependency,
    PackageNamespaceVersionDependency,
    PackageVersionIdDependency,
    StaticDependency,
    parse_dependencies,
)
from cumulusci.core.dependencies.github import (
    get_package_data,
    get_package_details_from_tag,
    get_remote_project_config,
    get_repo,
)
from cumulusci.core.exceptions import CumulusCIException, DependencyResolutionError
from cumulusci.core.github import (
    find_latest_release,
    find_repo_2gp_context,
    find_repo_feature_prefix,
    get_version_id_from_commit,
)
from cumulusci.core.versions import PackageType
from cumulusci.utils.git import (
    construct_release_branch_name,
    get_feature_branch_name,
    get_release_identifier,
    is_release_branch_or_child,
)


class DependencyResolutionStrategy(str, Enum):
    """Enum that defines a strategy for resolving a dynamic dependency into a static dependency."""

    STATIC_TAG_REFERENCE = "tag"
    COMMIT_STATUS_EXACT_BRANCH = "commit_status_exact_branch"
    COMMIT_STATUS_RELEASE_BRANCH = "commit_status_release_branch"
    COMMIT_STATUS_PREVIOUS_RELEASE_BRANCH = "commit_status_previous_release_branch"
    BETA_RELEASE_TAG = "latest_beta"
    RELEASE_TAG = "latest_release"
    UNMANAGED_HEAD = "unmanaged"


class Resolver(abc.ABC):
    """Abstract base class for dependency resolution strategies."""

    name = "Resolver"

    @abc.abstractmethod
    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        pass

    @abc.abstractmethod
    def resolve(
        self, dep: DynamicDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        pass

    def __str__(self):
        return self.name


class GitHubTagResolver(Resolver):
    """Resolver that identifies a ref by a specific GitHub tag."""

    name = "GitHub Tag Resolver"

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return isinstance(dep, BaseGitHubDependency) and dep.tag is not None

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        try:
            # Find the github release corresponding to this tag.
            repo = get_repo(dep.github, context)
            release = repo.release_from_tag(dep.tag)
            tag = repo.tag(repo.ref(f"tags/{release.tag_name}").object.sha)
            ref = tag.object.sha
            package_config = get_remote_project_config(repo, ref)
            package_name, namespace = get_package_data(package_config)
            version_id, package_type = get_package_details_from_tag(tag)

            install_unmanaged = (
                dep.is_unmanaged  # We've been told to use this dependency unmanaged
                or not (
                    # We will install managed if:
                    namespace  # the package has a namespace
                    or version_id  # or is a non-namespaced Unlocked Package
                )
            )

            if install_unmanaged:
                return ref, None
            else:
                if package_type is PackageType.SECOND_GEN:
                    package_dep = PackageVersionIdDependency(
                        version_id=version_id,
                        version_number=release.name,
                        package_name=package_name,
                    )
                else:
                    package_dep = PackageNamespaceVersionDependency(
                        namespace=namespace,
                        version=release.name,
                        package_name=package_name,
                        version_id=version_id,
                    )

                return (ref, package_dep)
        except NotFoundError:
            raise DependencyResolutionError(f"No release found for tag {dep.tag}")


class GitHubReleaseTagResolver(Resolver):
    """Resolver that identifies a ref by finding the latest GitHub release."""

    name = "GitHub Release Resolver"
    include_beta = False

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return isinstance(dep, BaseGitHubDependency)

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        repo = get_repo(dep.github, context)
        release = find_latest_release(repo, include_beta=self.include_beta)
        if release:
            tag = repo.tag(repo.ref(f"tags/{release.tag_name}").object.sha)
            version_id, package_type = get_package_details_from_tag(tag)

            ref = tag.object.sha
            package_config = get_remote_project_config(repo, ref)
            package_name, namespace = get_package_data(package_config)

            install_unmanaged = (
                dep.is_unmanaged  # We've been told to use this dependency unmanaged
                or not (
                    # We will install managed if:
                    namespace  # the package has a namespace
                    or version_id  # or is a non-namespaced Unlocked Package
                )
            )

            if install_unmanaged:
                return ref, None
            else:
                if package_type is PackageType.SECOND_GEN:
                    package_dep = PackageVersionIdDependency(
                        version_id=version_id,
                        version_number=release.name,
                        package_name=package_name,
                    )
                else:
                    package_dep = PackageNamespaceVersionDependency(
                        namespace=namespace,
                        version=release.name,
                        package_name=package_name,
                        version_id=version_id,
                    )
                return (ref, package_dep)

        return (None, None)


class GitHubBetaReleaseTagResolver(GitHubReleaseTagResolver):
    """Resolver that identifies a ref by finding the latest GitHub release, including betas."""

    name = "GitHub Release Resolver (Betas)"
    include_beta = True


class GitHubUnmanagedHeadResolver(Resolver):
    """Resolver that identifies a ref by finding the latest commit on the main branch."""

    name = "GitHub Unmanaged Resolver"

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return isinstance(dep, BaseGitHubDependency)

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        repo = get_repo(dep.github, context)
        return (repo.branch(repo.default_branch).commit.sha, None)


class GitHubReleaseBranchResolver(Resolver, abc.ABC):
    """Abstract base class for resolvers that use commit statuses on release branches to find refs."""

    def can_resolve(self, dep: DynamicDependency, context: BaseProjectConfig) -> bool:
        return self.is_valid_repo_context(context) and isinstance(
            dep, BaseGitHubDependency
        )

    def get_release_id(self, context: BaseProjectConfig) -> int:
        if not context.repo_branch or not context.project__git__prefix_feature:
            raise DependencyResolutionError(
                "Cannot get current branch or feature branch prefix"
            )
        release_id = get_release_identifier(
            context.repo_branch, context.project__git__prefix_feature
        )
        if not release_id:
            raise DependencyResolutionError("Cannot get current release identifier")

        return int(release_id)

    def is_valid_repo_context(self, context) -> bool:
        return (
            context.repo_branch
            and context.project__git__prefix_feature
            and is_release_branch_or_child(
                context.repo_branch, context.project__git__prefix_feature
            )
        )

    def locate_commit_status_package_id(self, remote_repo, release_branch, context_2gp):
        version_id = None
        count = 0
        commit = release_branch.commit
        while version_id is None and count < 5:
            version_id = get_version_id_from_commit(
                remote_repo, commit.sha, context_2gp
            )
            if version_id:
                break
            count += 1
            if commit.parents:
                commit = remote_repo.commit(commit.parents[0]["sha"])
            else:
                commit = None
                break

        return version_id, commit


class GitHubReleaseBranchCommitStatusResolver(GitHubReleaseBranchResolver):
    """Resolver that identifies a ref by finding a beta package version
    in a commit status on a `feature/NNN` release branch."""

    name = "GitHub Release Branch Commit Status Resolver"
    branch_depth = 1

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:

        release_id = self.get_release_id(context)
        repo = context.get_repo_from_url(dep.github)
        if not repo:
            raise DependencyResolutionError(
                f"Unable to access GitHub repository for {dep.github}"
            )

        try:
            remote_branch_prefix = find_repo_feature_prefix(repo)
            remote_2gp_context = find_repo_2gp_context(repo)
        except Exception:
            context.logger.info(
                f"Could not find feature branch prefix or commit-status context for {repo.clone_url}. Unable to resolve packages."
            )
            return (None, None)

        # We will check at least the release branch corresponding to our release id.
        # We may be configured to check backwards on release branches.
        release_branch = None
        for i in range(0, self.branch_depth):
            try:
                remote_matching_branch = construct_release_branch_name(
                    remote_branch_prefix, str(release_id - i)
                )

                release_branch = repo.branch(remote_matching_branch)
            except NotFoundError:
                pass

        if release_branch:
            version_id, commit = self.locate_commit_status_package_id(
                repo,
                release_branch,
                remote_2gp_context,
            )

            if version_id:
                context.logger.info(
                    f"Located package version {version_id} for release {release_id} on {repo.clone_url} at commit {release_branch.commit.sha}"
                )
                package_config = get_remote_project_config(repo, commit.sha)
                package_name, _ = get_package_data(package_config)

                return commit.sha, PackageVersionIdDependency(
                    version_id=version_id, package_name=package_name
                )

        context.logger.warn(
            f"No package version located for release {release_id} on {repo.clone_url}."
        )
        return (None, None)


class GitHubPreviousReleaseBranchCommitStatusResolver(
    GitHubReleaseBranchCommitStatusResolver
):
    """Resolver that identifies a ref by finding a beta package version
    in a commit status on a `feature/NNN` release branch that is earlier
    than the matching local release branch."""

    name = "GitHub Previous Release Branch Commit Status Resolver"
    branch_depth = 3


class GitHubReleaseBranchExactMatchCommitStatusResolver(GitHubReleaseBranchResolver):
    """Resolver that identifies a ref by finding a beta package version
    in a commit status on a branch whose name matches the local branch."""

    name = "GitHub Exact-Match Commit Status Resolver"

    def resolve(
        self, dep: BaseGitHubDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        release_id = self.get_release_id(context)

        repo = context.get_repo_from_url(dep.github)
        if not repo:
            raise DependencyResolutionError(
                f"Unable to access GitHub repository for {dep.github}"
            )

        try:
            remote_branch_prefix = find_repo_feature_prefix(repo)
            remote_2gp_context = find_repo_2gp_context(repo)
        except Exception:
            context.logger.info(
                f"Could not find feature branch prefix or commit-status context for {repo.clone_url}. Unable to resolve package."
            )
            return (None, None)

        # Attempt exact match
        try:
            branch = get_feature_branch_name(
                context.repo_branch, context.project__git__prefix_feature
            )
            release_branch = repo.branch(f"{remote_branch_prefix}{branch}")
        except Exception:
            context.logger.info(f"Exact-match branch not found for {repo.clone_url}.")
            return (None, None)

        version_id, commit = self.locate_commit_status_package_id(
            repo,
            release_branch,
            remote_2gp_context,
        )

        if version_id:
            context.logger.info(
                f"Located package version {version_id} for release {release_id} on {repo.clone_url} at commit {release_branch.commit.sha}"
            )

            package_config = get_remote_project_config(repo, commit.sha)
            package_name, _ = get_package_data(package_config)

            return commit.sha, PackageVersionIdDependency(
                version_id=version_id, package_name=package_name
            )

        context.logger.warn(
            f"No package version located for release {release_id} on {repo.clone_url}."
        )
        return (None, None)


RESOLVER_CLASSES = {
    DependencyResolutionStrategy.STATIC_TAG_REFERENCE: GitHubTagResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_EXACT_BRANCH: GitHubReleaseBranchExactMatchCommitStatusResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_RELEASE_BRANCH: GitHubReleaseBranchCommitStatusResolver,
    DependencyResolutionStrategy.COMMIT_STATUS_PREVIOUS_RELEASE_BRANCH: GitHubPreviousReleaseBranchCommitStatusResolver,
    DependencyResolutionStrategy.BETA_RELEASE_TAG: GitHubBetaReleaseTagResolver,
    DependencyResolutionStrategy.RELEASE_TAG: GitHubReleaseTagResolver,
    DependencyResolutionStrategy.UNMANAGED_HEAD: GitHubUnmanagedHeadResolver,
}


## External API


def get_resolver(
    strategy: DependencyResolutionStrategy, dependency: DynamicDependency
) -> Optional[Resolver]:
    """Return an instance of a resolver class capable of applying the specified
    resolution strategy to the dependency."""
    # This will be fleshed out when further types of DynamicDependency are added.

    return RESOLVER_CLASSES[strategy]()


def get_resolver_stack(
    context: BaseProjectConfig, name: str
) -> List[DependencyResolutionStrategy]:
    """Return a sequence of resolution strategies identified by the given `name`,
    which can be either a named strategy from project__dependency_resolutions__resolution_strategies
    or an alias like `production`."""
    resolutions = context.project__dependency_resolutions
    stacks = context.project__dependency_resolutions__resolution_strategies

    if name in resolutions and name != "resolution_strategies":
        name = resolutions[name]

    if stacks and name in stacks:
        return [DependencyResolutionStrategy(n) for n in stacks[name]]

    raise CumulusCIException(f"Resolver stack {name} was not found.")


def dependency_filter_ignore_deps(ignore_deps: List[dict]) -> Callable:
    ignore_github = [d["github"] for d in ignore_deps if "github" in d]
    ignore_namespace = [d["namespace"] for d in ignore_deps if "namespace" in d]

    def should_include(some_dep: Dependency) -> bool:

        if isinstance(some_dep, PackageNamespaceVersionDependency):
            return some_dep.namespace not in ignore_namespace
        if isinstance(some_dep, BaseGitHubDependency):
            return some_dep.github not in ignore_github

        return True

    return should_include


def get_static_dependencies(
    context: BaseProjectConfig,
    dependencies: List[Dependency] = None,
    resolution_strategy: str = None,
    strategies: List[DependencyResolutionStrategy] = None,
    filter_function: Callable = None,
) -> List[StaticDependency]:
    """Resolves the dependencies of a CumulusCI project
    to convert dynamic GitHub dependencies into static dependencies
    by inspecting the referenced repositories.

    Keyword arguments:
    :param context: a project config
    :param dependencies: an optional list of dependencies to resolve
                         (defaults to project__dependencies from the project_config)
    :param resolution_strategy: name of a resolution strategy or stack
    :param strategies: list of resolution strategies to use
                       (specify this or resolution_strategy but not both)
    :param filter_function: if provided, call the function with each dependency
                            (including transitive ones) encountered, and include
                            those for which True is returned.
    """
    if dependencies is None:
        dependencies = parse_dependencies(context.project__dependencies)
    assert (resolution_strategy is None and strategies is not None) or (
        strategies is None and resolution_strategy is not None
    ), "Expected resolution_strategy or strategies but not both"
    if resolution_strategy:
        strategies = get_resolver_stack(context, resolution_strategy)
    if filter_function is None:
        filter_function = lambda x: True  # noqa: E731

    while any(not d.is_flattened or not d.is_resolved for d in dependencies):
        for d in dependencies:
            if isinstance(d, DynamicDependency) and not d.is_resolved:
                d.resolve(context, strategies)

        def unique(it: Iterable):
            seen = set()

            for each in it:
                if each not in seen:
                    seen.add(each)
                    yield each

        dependencies = list(
            unique(
                itertools.chain(
                    *list(
                        d.flatten(context) for d in dependencies if filter_function(d)
                    )
                ),
            )
        )

    # Make sure, if we had no flattening or resolving to do, that we apply the ignore list.
    return [d for d in dependencies if filter_function(d)]


def resolve_dependency(
    dependency: DynamicDependency,
    context: BaseProjectConfig,
    strategies: List[DependencyResolutionStrategy],
):
    """Resolve a DynamicDependency that is not pinned to a specific version into one that is.

    If successful, sets `dependency.ref` and optionally `dependency.package_dependency`
    (if a package release is found).

    Otherwise raises DependencyResolutionError.
    """

    if dependency.is_resolved:
        return

    for s in strategies:
        resolver = get_resolver(s, dependency)

        if resolver and resolver.can_resolve(dependency, context):
            try:
                dependency.ref, dependency.package_dependency = resolver.resolve(
                    dependency, context
                )
                if dependency.package_dependency:
                    try:
                        dependency.package_dependency.password_env_name = (
                            dependency.password_env_name
                        )
                    except AttributeError:  # pragma: no cover
                        pass
                if dependency.ref:
                    break
            except DependencyResolutionError:
                context.logger.info(
                    f"Resolution strategy {s} failed for dependency {dependency}."
                )

    if not dependency.ref:
        raise DependencyResolutionError(f"Unable to resolve dependency {dependency}")



import pytest
import responses

from cumulusci.core.config.org_config import OrgConfig
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.tasks.salesforce.enable_prediction import EnablePrediction
from cumulusci.tests.util import DummyKeychain

from .util import create_task


@pytest.fixture
def task():
    return create_task(
        EnablePrediction,
        {"api_names": ["test_prediction_v0", "test_prediction_2_v0"]},
        org_config=OrgConfig(
            {"instance_url": "https://test-dev-ed.my.salesforce.com"},
            "test",
            keychain=DummyKeychain(),
        ),
    )


@pytest.fixture
def mock_oauth():
    with responses.RequestsMock() as rsps:
        rsps.add(
            "POST",
            "https://test-dev-ed.my.salesforce.com/services/oauth2/token",
            json={
                "access_token": "TOKEN",
                "instance_url": "https://test-dev-ed.my.salesforce.com",
            },
        )
        rsps.add(
            "GET",
            url="https://test-dev-ed.my.salesforce.com/services/oauth2/userinfo",
            json={},
            status=200,
        )
        rsps.add(
            "GET",
            url="https://test-dev-ed.my.salesforce.com/services/data",
            json=[
                {
                    "label": "Summer '21",
                    "url": "/services/data/v52.0",
                    "version": "52.0",
                }
            ],
            status=200,
        )
        rsps.add(
            "GET",
            "https://test-dev-ed.my.salesforce.com/services/data/v52.0/sobjects/Organization/",
            json={
                "OrganizationType": "Developer",
                "IsSandbox": False,
                "InstanceName": "NA149",
                "NamespacePrefix": None,
            },
        )

        yield rsps


def test_run_task(mock_oauth, task):
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT+Id+FROM+MLPredictionDefinition+WHERE+DeveloperName+%3D+%27test_prediction_v0%27",
        json={"totalSize": 1, "records": [{"Id": "001"}]},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT+Id+FROM+MLPredictionDefinition+WHERE+DeveloperName+%3D+%27test_prediction_2_v0%27",
        json={"totalSize": 1, "records": [{"Id": "002"}]},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/001",
        json={"Metadata": {"status": "Draft"}},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/002",
        json={"Metadata": {"status": "Draft"}},
    )
    mock_oauth.add(
        method="PATCH",
        url="https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/001",
        match=[responses.json_params_matcher({"Metadata": {"status": "Enabled"}})],
    )
    mock_oauth.add(
        method="PATCH",
        url="https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/002",
        match=[responses.json_params_matcher({"Metadata": {"status": "Enabled"}})],
    )

    task()


def test_run_task__not_found_exception(mock_oauth, task):
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT+Id+FROM+MLPredictionDefinition+WHERE+DeveloperName+%3D+%27test_prediction_v0%27",
        json={"totalSize": 0, "records": []},
    )

    with pytest.raises(CumulusCIException) as e:
        task()
        assert "not found" in str(e)


def test_run_task__failed_update_exception(mock_oauth, task):
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT+Id+FROM+MLPredictionDefinition+WHERE+DeveloperName+%3D+%27test_prediction_v0%27",
        json={"totalSize": 1, "records": [{"Id": "001"}]},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/001",
        json={"Metadata": {"status": "Draft"}},
    )
    mock_oauth.add(
        method="PATCH",
        url="https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/001",
        status=400,
    )

    with pytest.raises(CumulusCIException):
        task()


def test_run_task__namespaced_org(mock_oauth, task):
    task.options["namespaced_org"] = True
    task.options["namespace_inject"] = "foo"
    task.options["api_names"] = [
        "%%%NAMESPACED_ORG%%%test_prediction_v0",
        "%%%NAMESPACED_ORG%%%test_prediction_2_v0",
    ]

    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT+Id+FROM+MLPredictionDefinition+WHERE+DeveloperName+%3D+%27foo__test_prediction_v0%27",
        json={"totalSize": 1, "records": [{"Id": "001"}]},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT+Id+FROM+MLPredictionDefinition+WHERE+DeveloperName+%3D+%27foo__test_prediction_2_v0%27",
        json={"totalSize": 1, "records": [{"Id": "002"}]},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/001",
        json={"Metadata": {"status": "Draft"}},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/002",
        json={"Metadata": {"status": "Draft"}},
    )
    mock_oauth.add(
        method="PATCH",
        url="https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/001",
        match=[responses.json_params_matcher({"Metadata": {"status": "Enabled"}})],
    )
    mock_oauth.add(
        method="PATCH",
        url="https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/002",
        match=[responses.json_params_matcher({"Metadata": {"status": "Enabled"}})],
    )

    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT%20SubscriberPackage.Id,%20SubscriberPackage.NamespacePrefix,%20SubscriberPackageVersionId%20FROM%20InstalledSubscriberPackage",
        json={"totalSize": 0, "records": []},
    )
    mock_oauth.replace(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/sobjects/Organization/",
        json={
            "OrganizationType": "Developer",
            "IsSandbox": False,
            "InstanceName": "NA149",
            "NamespacePrefix": "foo",
        },
    )

    task()


def test_run_task__managed_org(mock_oauth, task):
    task.options["managed"] = True
    task.options["namespace_inject"] = "foo"
    task.options["api_names"] = [
        "%%%NAMESPACE%%%test_prediction_v0",
        "%%%NAMESPACE%%%test_prediction_2_v0",
    ]

    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT+Id+FROM+MLPredictionDefinition+WHERE+DeveloperName+%3D+%27foo__test_prediction_v0%27",
        json={"totalSize": 1, "records": [{"Id": "001"}]},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/query/?q=SELECT+Id+FROM+MLPredictionDefinition+WHERE+DeveloperName+%3D+%27foo__test_prediction_2_v0%27",
        json={"totalSize": 1, "records": [{"Id": "002"}]},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/001",
        json={"Metadata": {"status": "Draft"}},
    )
    mock_oauth.add(
        "GET",
        "https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/002",
        json={"Metadata": {"status": "Draft"}},
    )
    mock_oauth.add(
        method="PATCH",
        url="https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/001",
        match=[responses.json_params_matcher({"Metadata": {"status": "Enabled"}})],
    )
    mock_oauth.add(
        method="PATCH",
        url="https://test-dev-ed.my.salesforce.com/services/data/v52.0/tooling/sobjects/MLPredictionDefinition/002",
        match=[responses.json_params_matcher({"Metadata": {"status": "Enabled"}})],
    )

    task()



import functools
import io
import os
import re
import time
import webbrowser
from typing import Callable, Optional, Union
from urllib.parse import urlparse

import github3
from github3 import GitHub, login
from github3.exceptions import AuthenticationFailed, ResponseError, TransportError
from github3.git import Reference, Tag
from github3.pulls import ShortPullRequest
from github3.repos.commit import RepoCommit
from github3.repos.repo import Repository
from github3.session import GitHubSession
from requests.adapters import HTTPAdapter
from requests.exceptions import RetryError
from requests.models import Response
from requests.packages.urllib3.util.retry import Retry
from rich.console import Console

from cumulusci.core.exceptions import (
    DependencyLookupError,
    GithubApiError,
    GithubApiNotFoundError,
    GithubException,
)
from cumulusci.oauth.client import (
    OAuth2ClientConfig,
    OAuth2DeviceConfig,
    get_device_code,
    get_device_oauth_token,
)
from cumulusci.utils.http.requests_utils import safe_json_from_response
from cumulusci.utils.yaml.cumulusci_yml import cci_safe_load

OAUTH_DEVICE_APP = {
    "client_id": "2a4bc3e5ce4f2c49a957",
    "auth_uri": "https://github.com/login/device/code",
    "token_uri": "https://github.com/login/oauth/access_token",
    "scope": "repo gist",
}
SSO_WARNING = """Results may be incomplete. You have not granted your Personal Access token access to the following organizations:"""
UNAUTHORIZED_WARNING = """
Bad credentials. Verify that your personal access token is correct and that you are authorized to access this resource.
"""

# Prepare request retry policy to be attached to github sessions.
# 401 is a weird status code to retry, but sometimes it happens spuriously
# and https://github.community/t5/GitHub-API-Development-and/Random-401-errors-after-using-freshly-generated-installation/m-p/22905 suggests retrying
retries = Retry(status_forcelist=(401, 502, 503, 504), backoff_factor=0.3)
adapter = HTTPAdapter(max_retries=retries)


def get_github_api(username=None, password=None):
    """Old API that only handles logging in as a user.

    Here for backwards-compatibility during the transition.
    """
    gh = login(username, password)
    gh.session.mount("http://", adapter)
    gh.session.mount("https://", adapter)
    return gh


INSTALLATIONS = {}


def get_github_api_for_repo(keychain, owner, repo, session=None):
    gh = GitHub(
        session=session
        or GitHubSession(default_read_timeout=30, default_connect_timeout=30)
    )
    # Apply retry policy
    gh.session.mount("http://", adapter)
    gh.session.mount("https://", adapter)

    GITHUB_TOKEN = os.environ.get("GITHUB_TOKEN")
    APP_KEY = os.environ.get("GITHUB_APP_KEY", "").encode("utf-8")
    APP_ID = os.environ.get("GITHUB_APP_ID")
    if APP_ID and APP_KEY:
        installation = INSTALLATIONS.get((owner, repo))
        if installation is None:
            gh.login_as_app(APP_KEY, APP_ID, expire_in=120)
            try:
                installation = gh.app_installation_for_repository(owner, repo)
            except github3.exceptions.NotFoundError:
                raise GithubException(
                    f"Could not access {owner}/{repo} using GitHub app. "
                    "Does the app need to be installed for this repository?"
                )
            INSTALLATIONS[(owner, repo)] = installation
        gh.login_as_app_installation(APP_KEY, APP_ID, installation.id)
    elif GITHUB_TOKEN:
        gh.login(token=GITHUB_TOKEN)
    else:
        github_config = keychain.get_service("github")
        token = github_config.password or github_config.token
        gh.login(github_config.username, token)
    return gh


def validate_service(options: dict) -> dict:
    username = options["username"]
    token = options["token"]
    # Don't make the user wait 4 minutes to fail
    gh = GitHub(token=token)

    try:
        authed_user = gh.me()
        auth_login = authed_user.login
        assert username == auth_login, f"{username}, {auth_login}"
    except AssertionError as e:
        raise GithubException(
            f"Service username and token username do not match. ({str(e)})"
        )
    except Exception as e:
        warning_msg = format_github3_exception(e) or str(e)
        raise GithubException(
            f"Could not confirm access to the GitHub API: {warning_msg}"
        )
    else:
        member_orgs = {f"{org.id}": f"{org.login}" for org in gh.organizations()}
        options["Organizations"] = ", ".join([k for k in member_orgs.values()])

        # We're checking for a partial-response SSO header and /user/orgs
        # doesn't include one, so we need /user/repos instead.
        repo_generator = gh.repositories()
        _ = next(repo_generator, None)
        repo_response = repo_generator.last_response
        options["scopes"] = get_oauth_scopes(repo_response)

        unauthorized_org_ids = get_sso_disabled_orgs(repo_response)
        unauthorized_orgs = {
            k: member_orgs[k] for k in unauthorized_org_ids if k in member_orgs
        }
        if unauthorized_orgs:
            options["SSO Disabled"] = ", ".join([k for k in unauthorized_orgs.values()])

        expiration_date = repo_response.headers.get(
            "GitHub-Authentication-Token-Expiration"
        )
        if expiration_date:
            options["expires"] = expiration_date

    return options


def get_pull_requests_with_base_branch(repo, base_branch_name, head=None, state=None):
    """Returns a list of pull requests with the given base branch"""
    if head:
        head = repo.owner.login + ":" + head
    return list(repo.pull_requests(base=base_branch_name, head=head, state=state))


def get_pull_requests_by_head(repo, branch_name):
    """Returns all pull requests with head equal to the given branch name."""
    if branch_name == repo.default_branch:
        return None

    return list(repo.pull_requests(head=repo.owner.login + ":" + branch_name))


def create_pull_request(repo, branch_name, base=None, title=None):
    """Creates a pull request for the given branch"""
    base = base or repo.default_branch
    title = title or "Auto-Generated Pull Request"
    pull_request = repo.create_pull(title, base, branch_name)
    return pull_request


def add_labels_to_pull_request(repo, pull_request, *labels):
    """Adds a label to a pull request via the issue object
    Args:
    * repo: Repository object
    * pull_request: ShortPullRequest object that exists in repo
    * labels: list(str) of labels to add to the pull request"""
    issue = repo.issue(pull_request.number)
    issue.add_labels(*labels)


def is_label_on_pull_request(repo, pull_request, label_name):
    """Returns True if the given label is on the pull request with the given
    pull request number. False otherwise."""
    labels = list(repo.issue(pull_request.number).labels())
    return any(label_name == issue_label.name for issue_label in labels)


def get_pull_requests_by_commit(github, repo, commit_sha):
    endpoint = (
        github.session.base_url
        + f"/repos/{repo.owner.login}/{repo.name}/commits/{commit_sha}/pulls"
    )
    response = github.session.get(
        endpoint, headers={"Accept": "application/vnd.github.groot-preview+json"}
    )
    json_list = safe_json_from_response(response)

    # raises github3.exceptions.IncompleteResposne
    # when these are not present
    for json in json_list:
        json["body_html"] = ""
        json["body_text"] = ""

    return [ShortPullRequest(json, github) for json in json_list]


def is_pull_request_merged(pull_request):
    """Takes a github3.pulls.ShortPullRequest object"""
    return pull_request.merged_at is not None


def markdown_link_to_pr(change_note):
    return f"{change_note.title} [[PR{change_note.number}]({change_note.html_url})]"


def find_latest_release(repo, include_beta=None):
    try:
        if include_beta:
            return next(repo.releases())
        else:
            return repo.latest_release()
    except (github3.exceptions.NotFoundError, StopIteration):
        pass


def find_previous_release(repo, prefix=None):
    most_recent = None
    for release in repo.releases():
        if prefix and not release.tag_name.startswith(prefix):
            continue
        if not prefix and release.prerelease:
            continue
        # Return the second release
        if most_recent is None:
            most_recent = release
        else:
            return release


def create_gist(github, description, files):
    """Creates a gist with the given description and files.

    github - an
    description - str
    files - A dict of files in the form of {filename:{'content': content},...}
    """
    return github.create_gist(description, files, public=False)


VERSION_ID_RE = re.compile(r"version_id: (\S+)")


def get_version_id_from_commit(repo, commit_sha, context):
    commit = get_commit(repo, commit_sha)

    for status in commit.status().statuses:
        if status.state == "success" and status.context == context:
            match = VERSION_ID_RE.search(status.description)
            if match:
                return match.group(1)


def get_commit(repo: Repository, commit_sha: str) -> Optional[RepoCommit]:
    """Given a SHA1 hash, retrieve a Commit object from the REST API."""
    try:
        commit = repo.commit(commit_sha)
    except (github3.exceptions.NotFoundError, github3.exceptions.UnprocessableEntity):
        # GitHub returns 422 for nonexistent commits in at least some circumstances.
        raise DependencyLookupError(f"Could not find commit {commit_sha} on GitHub")
    return commit


def find_repo_feature_prefix(repo: Repository) -> str:
    contents = repo.file_contents(
        "cumulusci.yml",
        ref=repo.branch(repo.default_branch).commit.sha,
    )
    head_cumulusci_yml = cci_safe_load(io.StringIO(contents.decoded.decode("utf-8")))
    return (
        head_cumulusci_yml.get("project", {})
        .get("git", {})
        .get("prefix_feature", "feature/")
    )


def find_repo_2gp_context(repo: Repository) -> str:
    contents = repo.file_contents(
        "cumulusci.yml",
        ref=repo.branch(repo.default_branch).commit.sha,
    )
    head_cumulusci_yml = cci_safe_load(io.StringIO(contents.decoded.decode("utf-8")))
    return (
        head_cumulusci_yml.get("project", {})
        .get("git", {})
        .get(
            "2gp_context", "Build Feature Test Package"
        )  # TODO: source default from our `cumulusci.yml`
    )


def get_tag_by_name(repo: Repository, tag_name: str) -> Tag:
    """Fetches a tag by name from the given repository"""
    ref: Reference = get_ref_for_tag(repo, tag_name)
    try:
        return repo.tag(ref.object.sha)
    except github3.exceptions.NotFoundError:
        msg = f"Could not find tag '{tag_name}' with SHA {ref.object.sha} on GitHub"
        if ref.object.type != "tag":
            msg += f"\n{tag_name} is not an annotated tag."
        raise GithubApiNotFoundError(msg)


def get_ref_for_tag(repo: Repository, tag_name: str) -> Reference:
    """Gets a Reference object for the tag with the given name"""
    try:
        return repo.ref(f"tags/{tag_name}")
    except github3.exceptions.NotFoundError:
        raise GithubApiNotFoundError(
            f"Could not find reference for 'tags/{tag_name}' on GitHub"
        )


def get_version_id_from_tag(repo: Repository, tag_name: str) -> str:
    """Given the name of a tag, return the version_id in the tag's message.

    @param tag_name: the name of the tag
    @param repo: the repository of the package to look for a release in
    @returns: the 04t id in the tag's messages
    """
    tag = get_tag_by_name(repo, tag_name)
    for line in tag.message.split("\n"):
        if line.startswith("version_id:"):
            version_id = line.split("version_id: ")[1]
            if not version_id.startswith("04t"):
                continue
            return version_id

    raise DependencyLookupError(f"Could not find version_id for tag {tag_name}")


def format_github3_exception(exc: Union[ResponseError, TransportError]) -> str:
    """Checks github3 exceptions for the most common GitHub authentication
    issues, returning a user-friendly message if found.

    @param exc: The exception to process
    @returns: The formatted exception string
    """
    user_warning = ""

    too_many_str = "too many 401 error responses"
    is_bad_auth_retry = (
        type(exc) is TransportError
        and type(exc.exception) is RetryError
        and too_many_str in str(exc.exception)
    )
    is_auth_failure = type(exc) is AuthenticationFailed

    if is_bad_auth_retry or is_auth_failure:
        user_warning = UNAUTHORIZED_WARNING

    if isinstance(exc, ResponseError):
        scope_error_msg = check_github_scopes(exc)
        sso_error_msg = check_github_sso_auth(exc)
        user_warning = scope_error_msg + sso_error_msg

    return user_warning


def warn_oauth_restricted(exc: ResponseError) -> str:
    user_warning = ""

    is_403 = exc.response.status_code == 403
    org_restricted_oauth_warning = (
        "organization has enabled OAuth App access restrictions"
    )

    if is_403 and org_restricted_oauth_warning in str(exc):
        user_warning = str(exc)
        user_warning += "\nYou may also use a Personal Access Token as a workaround."

    return user_warning


def check_github_scopes(exc: ResponseError) -> str:
    """
    Parse github3 ResponseError headers for the correct scopes and return a
    warning if the user is missing.

    @param exc: The exception to process
    @returns: The formatted exception string
    """

    user_warning = ""

    has_wrong_status_code = exc.response.status_code not in (403, 404)
    if has_wrong_status_code:
        return user_warning

    token_scopes = get_oauth_scopes(exc.response)

    # Gist resource won't return X-Accepted-OAuth-Scopes for some reason, so this
    # string might be `None`; we discard the empty string if so.
    accepted_scopes = exc.response.headers.get("X-Accepted-OAuth-Scopes") or ""
    accepted_scopes = set(accepted_scopes.split(", "))
    accepted_scopes.discard("")

    request_url = urlparse(exc.response.url)
    if not accepted_scopes and request_url.path == "/gists":
        accepted_scopes = {"gist"}

    missing_scopes = accepted_scopes.difference(token_scopes)
    if missing_scopes:
        user_warning = f"Your token may be missing the following scopes: {', '.join(missing_scopes)}\n"
        # This assumes we're not on enterprise and 'api.github.com' == request_url.hostname
        user_warning += (
            "Visit Settings > Developer settings > Personal access tokens to add them."
        )

    return user_warning


def check_github_sso_auth(exc: ResponseError) -> str:
    """
    Check ResponseError header for SSO authorization and return a warning if
    required

    @param exc: The exception to process
    @returns: The formatted exception string
    """
    user_warning = ""
    headers = exc.response.headers

    if exc.response.status_code != 403 or "X-Github-Sso" not in headers:
        return user_warning

    sso_header = str(headers["X-Github-Sso"] or "")
    if sso_header.startswith("required; url="):
        # In this case the message from github is good enough, but we can help
        # the user by opening a browser to authorize the token.
        auth_url = sso_header.split("url=", maxsplit=1)[1]
        user_warning = f"{exc.message}\n{auth_url}"
        webbrowser.open(auth_url)
    elif sso_header.startswith("partial-results"):
        # In cases where we don't have complete results we get the
        # partal-results header, so return the organization IDs. This may or
        # may not be useful without help from us to lookup the org IDs.
        unauthorized_org_ids = get_sso_disabled_orgs(exc.response)
        user_warning = f"{SSO_WARNING} {unauthorized_org_ids}"

    return user_warning


def get_sso_disabled_orgs(response: Response) -> list:
    """
    Given a response from Github, return a list of organization IDs without SSO
    grants.
    """
    disabled_orgs = []
    sso_header = response.headers.get("X-Github-Sso")
    partial_results_prefix = "partial-results; organizations="

    if sso_header and partial_results_prefix in sso_header:
        disabled_orgs = sso_header[len(partial_results_prefix) :].split(",")

    return disabled_orgs


def get_oauth_scopes(response: Response) -> set:
    """
    Given a response from Github, return the set of OAuth scopes for its
    request.
    """
    authorized_scopes = set()

    # If the token isn't authorized "X-OAuth-Scopes" header won't be present
    x_oauth_scopes = response.headers.get("X-OAuth-Scopes")
    if x_oauth_scopes:
        authorized_scopes = set(x_oauth_scopes.split(", "))

    return authorized_scopes


def catch_common_github_auth_errors(func: Callable) -> Callable:
    """
    A decorator catching the most common Github authentication errors.
    """

    @functools.wraps(func)
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (ResponseError, TransportError) as exc:
            error_msg = format_github3_exception(exc)
            if error_msg:
                raise GithubApiError(error_msg)
            else:
                raise

    return inner


def get_oauth_device_flow_token():
    """Interactive github authorization"""
    config = OAuth2ClientConfig(**OAUTH_DEVICE_APP)
    device_code = OAuth2DeviceConfig(**get_device_code(config))

    console = Console()
    console.print(
        f"[bold] Enter this one-time code: [red]{device_code.user_code}[/red][/bold]"
    )

    console.print(f"Opening {device_code.verification_uri} in your default browser...")
    webbrowser.open(device_code.verification_uri)
    time.sleep(2)  # Give the user a second or two before we start polling

    with console.status("Polling server for authorization..."):
        device_token: dict = get_device_oauth_token(
            client_config=config, device_config=device_code
        )

    access_token = device_token.get("access_token")
    if access_token:
        console.print(
            f"[bold green]Successfully authorized OAuth token ({access_token[:7]}...)[/bold green]"
        )

    return access_token



import copy

import pytest

from cumulusci.core import utils
from cumulusci.core.exceptions import ConfigMergeError, CumulusCIException


@pytest.fixture
def universal_config():
    return {
        "flows": {
            "steps_all_flows": {
                "group": "flowy",
                "description": "Elaborate description",
                "steps": {
                    1: {"flow": "short_flow", "options": {"foo": "bar"}},
                    2: {"flow": "mediuem_flow", "options": {"baz": "boop"}},
                    3: {"flow": "long_flow", "options": {"bim": "bap"}},
                },
            },
            "steps_all_tasks": {
                "group": "tasky",
                "description": "Elaborate description",
                "steps": {
                    1: {"task": "short_task", "options": {"foo": "bar"}},
                    2: {"task": "mediuem_task", "options": {"baz": "boop"}},
                    3: {"task": "long_task", "options": {"bim": "bap"}},
                },
            },
        }
    }


def test_init():
    config = utils.merge_config(
        {
            "universal_config": {"hello": "world"},
            "user_config": {"hello": "christian"},
        }
    )
    assert config["hello"] == "christian"


def test_merge_failure():
    with pytest.raises(ConfigMergeError) as cm:
        utils.merge_config(
            {
                "universal_config": {"hello": "world", "test": {"sample": 1}},
                "user_config": {"hello": "christian", "test": [1, 2]},
            }
        )
    exception = cm.value
    assert exception.config_name == "user_config"


def test_cleanup_flow_step_override_conflicts__no_op(universal_config):
    configs = {"universal_config": universal_config}
    clean_configs = utils.cleanup_flow_step_override_conflicts(configs)
    assert universal_config == clean_configs["universal_config"]


def test_cleanup_flow_step_override_conflicts__ambiguous_step(universal_config):
    project_config = {
        "flows": {"steps_all_flows": {"steps": {3: {"flow": "None", "task": "None"}}}}
    }

    configs = {"project_config": project_config, "universal_config": universal_config}
    with pytest.raises(CumulusCIException):
        utils.cleanup_flow_step_override_conflicts(configs)


def test_cleanup_flow_step_override_conflicts__task_overrides_flow(universal_config):
    project_config = {
        "flows": {
            "steps_all_flows": {
                "steps": {3: {"task": "custom_task", "options": {"super": "cool"}}}
            }
        }
    }
    # Copy things before they're operated on
    expected_universal_config = copy.deepcopy(universal_config)
    expected_universal_config["flows"]["steps_all_flows"]["steps"][3] = {}

    configs = {"project_config": project_config, "universal_config": universal_config}
    clean_configs = utils.cleanup_flow_step_override_conflicts(configs)

    assert expected_universal_config == clean_configs["universal_config"]


def test_cleanup_flow_step_override_conflicts__flow_overrides_task(universal_config):
    project_config = {
        "flows": {
            "steps_all_tasks": {
                "steps": {3: {"flow": "custom_task", "options": {"super": "cool"}}}
            }
        }
    }
    # Copy things before they're operated on
    expected_universal_config = copy.deepcopy(universal_config)
    expected_universal_config["flows"]["steps_all_tasks"]["steps"][3] = {}

    configs = {"project_config": project_config, "universal_config": universal_config}
    clean_configs = utils.cleanup_flow_step_override_conflicts(configs)

    assert expected_universal_config == clean_configs["universal_config"]


def test_cleanup_flow_step_override__old_syntax_task_to_flow(universal_config):
    """The 'old' syntax was to set the current step type to 'None' and the new
    step type with the name of the task/flow you want."""
    project_config = {
        "flows": {
            "steps_all_tasks": {
                "steps": {
                    3: {
                        "task": "None",
                        "flow": "custom_task",
                        "options": {"super": "cool"},
                    }
                }
            }
        }
    }
    # Copy things before they're operated on
    expected_project_config = copy.deepcopy(project_config)
    expected_project_config["flows"]["steps_all_tasks"]["steps"][3] = {
        # "task" should no longer be present
        "flow": "custom_task",
        "options": {"super": "cool"},
    }

    configs = {"project_config": project_config, "universal_config": universal_config}
    clean_configs = utils.cleanup_flow_step_override_conflicts(configs)

    assert expected_project_config == clean_configs["project_config"]


def test_cleanup_flow_step_override__old_syntax_flow_to_task(universal_config):
    """The 'old' syntax was to set the current step type to 'None' and the new
    step type with the name of the task/flow you want."""
    project_config = {
        "flows": {
            "steps_all_flows": {
                "steps": {
                    3: {
                        "flow": "None",
                        "task": "custom_task",
                        "options": {"super": "cool"},
                    }
                }
            }
        }
    }
    # Copy things before they're operated on
    expected_project_config = copy.deepcopy(project_config)
    expected_project_config["flows"]["steps_all_flows"]["steps"][3] = {
        # "task" should no longer be present
        "task": "custom_task",
        "options": {"super": "cool"},
    }

    configs = {"project_config": project_config, "universal_config": universal_config}
    clean_configs = utils.cleanup_flow_step_override_conflicts(configs)

    assert expected_project_config == clean_configs["project_config"]


def test_cleanup_flow_step_override_conflicts__multiple_overrides_of_alternating_types(
    universal_config,
):
    """This is to simulate a rare but plausible case where an user has multiple overrides of the same flow step
    accross three different config objects: global, project, and project_local.

    The result should be that any flow steps with differing types will be set to an empty dict, as
    was tested above. We also expect that any steps of the same type will _not_ be set to an empty dict.
    (Steps of the same type will override each other when dictmerge() is called)
    """
    # Throw in the old replace syntax just for fun
    project_local_config = {
        "flows": {
            "steps_all_tasks": {
                "steps": {
                    3: {
                        "task": "None",
                        "flow": "custom_flow_one",
                        "options": {"Trog": "dor"},
                    }
                }
            }
        }
    }
    project_config = {
        "flows": {
            "steps_all_tasks": {
                "steps": {3: {"task": "custom_task", "options": {"deploy": "clippy"}}}
            }
        }
    }
    # Even throwing in the old replace syntax just for fun
    global_config = {
        "flows": {
            "steps_all_tasks": {
                "steps": {
                    3: {
                        "flow": "custom_flow_two",
                        "options": {"Just": "do it"},
                    }
                }
            }
        }
    }

    # Copy things before they are operated on
    expected_universal_config = copy.deepcopy(universal_config)
    expected_universal_config["flows"]["steps_all_tasks"]["steps"][3] = {}

    expected_project_config = copy.deepcopy(project_config)
    expected_project_config["flows"]["steps_all_tasks"]["steps"][3] = {}

    expected_project_local_config = copy.deepcopy(project_local_config)
    expected_project_local_config["flows"]["steps_all_tasks"]["steps"][3] = {
        "flow": "custom_flow_one",
        "options": {"Trog": "dor"},
    }

    expected_global_config = copy.deepcopy(global_config)

    configs = {
        "global_config": global_config,
        "project_config": project_config,
        "project_local_config": project_local_config,
        "universal_config": universal_config,
    }
    clean_configs = utils.cleanup_flow_step_override_conflicts(configs)

    assert expected_universal_config == clean_configs["universal_config"]
    assert expected_project_config == clean_configs["project_config"]
    assert expected_project_local_config == clean_configs["project_local_config"]
    # This should remain unchanged
    assert expected_global_config == clean_configs["global_config"]



import pytest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.metadata_etl import AddValueSetEntries
from cumulusci.tasks.metadata_etl.value_sets import (
    CASE_STATUS_ERR,
    FULL_NAME_AND_LABEL_ERR,
    LEAD_STATUS_ERR,
    OPP_STAGE_ERR,
)
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.utils.xml import lxml_parse_string, metadata_tree

MD = "{%s}" % "http://soap.sforce.com/2006/04/metadata"
VALUESET_XML = b"""<?xml version="1.0" encoding="UTF-8"?>
<StandardValueSet xmlns="http://soap.sforce.com/2006/04/metadata">
    <sorted>false</sorted>
    <standardValue>
        <fullName>Value</fullName>
        <default>true</default>
        <label>Value</label>
    </standardValue>
    <standardValue>
        <fullName>Other</fullName>
        <default>false</default>
        <label>Other</label>
    </standardValue>
</StandardValueSet>
"""


class TestAddValueSetEntries:
    def test_adds_entry(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "entries": [
                    {"fullName": "Test", "label": "Label"},
                    {"fullName": "Test_2", "label": "Label 2"},
                ],
            },
        )

        tree = lxml_parse_string(VALUESET_XML)

        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Test']")) == 0
        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Test_2']")) == 0

        result = task._transform_entity(
            metadata_tree.fromstring(VALUESET_XML), "ValueSet"
        )

        entry = result._element.findall(f".//{MD}standardValue[{MD}fullName='Test']")
        assert len(entry) == 1
        label = entry[0].findall(f".//{MD}label")
        assert len(label) == 1
        assert label[0].text == "Label"
        default = entry[0].findall(f".//{MD}default")
        assert len(default) == 1
        assert default[0].text == "false"

        entry = result._element.findall(f".//{MD}standardValue[{MD}fullName='Test_2']")
        assert len(entry) == 1
        label = entry[0].findall(f".//{MD}label")
        assert len(label) == 1
        assert label[0].text == "Label 2"
        default = entry[0].findall(f".//{MD}default")
        assert len(default) == 1
        assert default[0].text == "false"

    def test_adds_entry__opportunitystage(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "OpportunityStage",
                "entries": [
                    {
                        "fullName": "Test",
                        "label": "Label",
                        "closed": True,
                        "won": True,
                        "forecastCategory": "Omitted",
                        "probability": 100,
                    }
                ],
            },
        )

        tree = metadata_tree.fromstring(VALUESET_XML)

        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Test']")) == 0
        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Test_2']")) == 0

        result = task._transform_entity(tree, "OpportunityStage")

        entry = result._element.findall(f".//{MD}standardValue[{MD}fullName='Test']")
        assert len(entry) == 1
        label = entry[0].findall(f".//{MD}label")
        assert len(label) == 1
        assert label[0].text == "Label"
        default = entry[0].findall(f".//{MD}default")
        assert len(default) == 1
        assert default[0].text == "false"
        closed = entry[0].findall(f".//{MD}closed")
        assert len(closed) == 1
        assert closed[0].text == "true"
        won = entry[0].findall(f".//{MD}won")
        assert len(won) == 1
        assert won[0].text == "true"
        forecastCategory = entry[0].findall(f".//{MD}forecastCategory")
        assert len(forecastCategory) == 1
        assert forecastCategory[0].text == "Omitted"
        probability = entry[0].findall(f".//{MD}probability")
        assert len(probability) == 1
        assert probability[0].text == "100"

    def test_adds_entry__casestatus(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "CaseStatus",
                "entries": [{"fullName": "Test", "label": "Label", "closed": True}],
            },
        )

        tree = lxml_parse_string(VALUESET_XML)

        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Test']")) == 0
        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Test_2']")) == 0

        result = task._transform_entity(
            metadata_tree.fromstring(VALUESET_XML), "CaseStatus"
        )

        entry = result._element.findall(f".//{MD}standardValue[{MD}fullName='Test']")
        assert len(entry) == 1
        label = entry[0].findall(f".//{MD}label")
        assert len(label) == 1
        assert label[0].text == "Label"
        default = entry[0].findall(f".//{MD}default")
        assert len(default) == 1
        assert default[0].text == "false"
        closed = entry[0].findall(f".//{MD}closed")
        assert len(closed) == 1
        assert closed[0].text == "true"

    def test_adds_entry__leadStatus(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "LeadStatus",
                "entries": [{"fullName": "Test", "label": "Label", "converted": True}],
            },
        )

        tree = lxml_parse_string(VALUESET_XML)

        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Test']")) == 0
        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Test_2']")) == 0

        result = task._transform_entity(
            metadata_tree.fromstring(VALUESET_XML), "LeadStatus"
        )

        entry = result._element.findall(f".//{MD}standardValue[{MD}fullName='Test']")
        assert len(entry) == 1
        label = entry[0].findall(f".//{MD}label")
        assert len(label) == 1
        assert label[0].text == "Label"
        default = entry[0].findall(f".//{MD}default")
        assert len(default) == 1
        assert default[0].text == "false"
        converted = entry[0].findall(f".//{MD}converted")
        assert len(converted) == 1
        assert converted[0].text == "true"

    def test_does_not_add_existing_entry(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "entries": [{"fullName": "Value", "label": "Label"}],
            },
        )

        tree = lxml_parse_string(VALUESET_XML)

        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Value']")) == 1

        metadata = metadata_tree.fromstring(VALUESET_XML)
        task._transform_entity(metadata, "ValueSet")

        assert len(tree.findall(f".//{MD}standardValue[{MD}fullName='Value']")) == 1

    entries = [
        {},
        {"fullName": "Value"},
        {"label": "Value"},
    ]

    @pytest.mark.parametrize("entry", entries)
    def test_raises_exception_missing_values(self, entry):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "entries": [entry],
            },
        )
        tree = metadata_tree.fromstring(VALUESET_XML)

        with pytest.raises(TaskOptionsError, match=FULL_NAME_AND_LABEL_ERR):
            task._transform_entity(tree, "ValueSet")

    def test_raises_exception_missing_values__opportunitystage(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "OpportunityStage",
                "entries": [{"fullName": "Value", "label": "Value"}],
            },
        )
        tree = metadata_tree.fromstring(VALUESET_XML)

        with pytest.raises(TaskOptionsError, match=OPP_STAGE_ERR):
            task._transform_entity(tree, "OpportunityStage")

    def test_raises_exception_missing_values__casestatus(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "CaseStatus",
                "entries": [{"fullName": "Value", "label": "Value"}],
            },
        )
        tree = metadata_tree.fromstring(VALUESET_XML)

        with pytest.raises(TaskOptionsError, match=CASE_STATUS_ERR):
            task._transform_entity(tree, "CaseStatus")

    def test_raises_exception_missing_values__leadstatus(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "LeadStatus",
                "entries": [{"fullName": "Value", "label": "Value"}],
            },
        )
        tree = metadata_tree.fromstring(VALUESET_XML)

        with pytest.raises(TaskOptionsError, match=LEAD_STATUS_ERR):
            task._transform_entity(tree, "LeadStatus")

    def test_adds_correct_number_of_values(self):
        task = create_task(
            AddValueSetEntries,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "entries": [
                    {"fullName": "Test", "label": "Label"},
                    {"fullName": "Test_2", "label": "Label 2"},
                    {"fullName": "Other", "label": "Duplicate"},
                ],
            },
        )

        mdtree = metadata_tree.fromstring(VALUESET_XML)
        xml_tree = mdtree._element

        assert len(xml_tree.findall(f".//{MD}standardValue")) == 2

        task._transform_entity(mdtree, "ValueSet")

        assert len(xml_tree.findall(f".//{MD}standardValue")) == 4



import fnmatch
import io
import os
import signal
import unittest
from unittest import mock

from selenium.common.exceptions import InvalidSelectorException

from cumulusci.tasks.robotframework import debugger


class TestDebugListener(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestDebugListener, cls).setUpClass()
        cls.listener = debugger.DebugListener()

    def test_listener_default_breakpoint(self):
        """Verify we get the correct default breakpoint"""
        listener = debugger.DebugListener()
        self.assertEqual(len(listener.breakpoints), 1)
        bp = listener.breakpoints[0]
        self.assertFalse(bp.temporary)
        self.assertEqual(
            bp.pattern, "*::cumulusci.robotframework.Salesforce.Breakpoint"
        )
        self.assertEqual(bp.breakpoint_type, debugger.Keyword)
        self.assertEqual(bp.regex, fnmatch.translate(bp.pattern))

    def test_listener_custom_breakpoints(self):
        """Verify we can create a cli with custom breakpoints"""
        breakpoints = [
            debugger.Breakpoint(debugger.Keyword, "*::keyword breakpoint"),
            debugger.Breakpoint(debugger.Testcase, "*::test breakpoint"),
            debugger.Breakpoint(debugger.Suite, "*::suite breakpoint"),
        ]
        listener = debugger.DebugListener(*breakpoints)
        self.assertEqual(listener.breakpoints, breakpoints)

    def test_listener_stack(self):
        """Verify that the listener properly tracks the stack as a test is executed"""
        self.listener.start_suite("Root", {"longname": "Root"})
        self.listener.start_suite("folder", {"longname": "Root.folder"})
        self.listener.start_suite("example", {"longname": "Root.folder.example"})
        self.listener.start_test("Test 1", {"longname": "Root.folder.example.Test 1"})
        self.listener.start_keyword("BuiltIn.Log", {"kwname": "Log"})

        # The listener stack should now include four elements, one for
        # each call to a listener method
        self.assertEqual(len(self.listener.stack), 5)

        self.assertEqual(self.listener.stack[0].name, "Root")
        self.assertEqual(self.listener.stack[1].name, "folder")
        self.assertEqual(self.listener.stack[2].name, "example")
        self.assertEqual(self.listener.stack[3].name, "Test 1")
        self.assertEqual(self.listener.stack[4].name, "BuiltIn.Log")

        # now, unwind the stack and make sure it's empty
        self.listener.end_keyword("BuiltIn.Log", {})
        self.listener.end_test("Test 1", {})
        self.listener.end_suite("example", {})
        self.listener.end_suite("folder", {})
        self.listener.end_suite("Root", {})
        self.assertEqual(len(self.listener.stack), 0)

    def test_listener_step(self):
        """Verify that the 'step' debugger command creates a breakpoint for the next step"""
        self.listener.start_suite("Root", {"longname": "Root"})
        self.listener.start_suite("example", {"longname": "Root.example"})
        self.listener.start_test("Test 1", {"longname": "Root.example.Test 1"})
        self.listener.start_keyword("cumulusci.Salesforce.breakpoint", {})

        self.assertEqual(
            len(self.listener.breakpoints),
            1,
            "Weird. There should have only been a single breakpoint",
        )
        # call `do_step` of the *listener*, not the debugger UI.
        # the debugger ui "do_step" method will both add a new breakpoint
        # and then continue to that breakpoint, and then that breakpoint
        # will be removed. This test is verifying only that the breakpoing
        # is added
        self.listener.do_step()

        # The 'step' command should cause a new temporary breakpoint to be added
        # in the same context as the current keyword.
        self.assertEqual(
            len(self.listener.breakpoints),
            2,
            "Expected a breakpoint to be added on the 'step' command",
        )
        self.assertEqual(
            self.listener.breakpoints[-1].pattern, "Root.example.Test 1::*"
        )
        self.assertTrue(self.listener.breakpoints[-1].temporary)

    def test_temporary_breakpoint(self):
        """Verify that a temporary breakpoint is removed when encountered"""
        bp1 = debugger.Breakpoint(debugger.Keyword, "*::breakpoint", temporary=False)
        bp2 = debugger.Breakpoint(
            debugger.Keyword, "*::temporary breakpoint", temporary=True
        )
        listener = debugger.DebugListener(bp1, bp2)
        listener.rdb = mock.Mock()
        # The listener uses the .intro attribute as a string, so
        # we need to set it to something so that the listener
        # doesnt' crash.
        listener.rdb.intro = "... intro ..."
        self.assertEqual(len(listener.breakpoints), 2)

        listener.start_suite("Suite", attrs={})
        listener.start_test("Test Case", attrs={})
        listener.start_keyword("temporary breakpoint", attrs={"args": ["one", "two"]})
        self.assertEqual(len(listener.breakpoints), 1)
        self.assertEqual(
            listener.breakpoints[0].pattern,
            "*::breakpoint",
            "the wrong breakpoint was removed",
        )
        listener.rdb.cmdloop.assert_called_once()


class TestRobotDebugger(unittest.TestCase):
    """These tests are for the DebuggerCli class

    This class has methods of the form 'do_<something'> where
    <something> is a debugger command (step, continue, vars, etc).

    Instead of actually running robot with the debugger, these tests
    simply verify that each of those functions returns the right thing
    and calls the right functions.

    """

    @classmethod
    def setUpClass(cls):
        super(TestRobotDebugger, cls).setUpClass()
        cls.mock_listener = mock.Mock()
        cls.mock_builtin = mock.Mock()

    def setUp(self):
        self.stdout = io.StringIO()
        self.cli = debugger.DebuggerCli(self.mock_listener, stdout=self.stdout)
        self.cli.builtin = self.mock_builtin

    def test_comment(self):
        """The comment command does nothing; we just need to verify
        that it returns None, which keeps the cmd REPL alive"""
        return_value = self.cli.default("# this is a comment")
        self.assertEqual(return_value, None)

    def test_variable_shortcuts(self):
        """Verify that just passing a variable name to the debugger fetches the value of that variable"""
        # scalar variable
        return_value = self.cli.default("${varname}")
        self.mock_builtin.get_variable_value.assert_called_with("${varname}")
        self.assertIsNone(return_value)

        # list variable
        return_value = self.cli.default("@{varname}")
        self.mock_builtin.get_variable_value.assert_called_with("@{varname}")
        self.assertIsNone(return_value)

        # dictionary variable
        return_value = self.cli.default("&{varname}")
        self.mock_builtin.get_variable_value.assert_called_with("&{varname}")
        self.assertIsNone(return_value)

        # None of the above. Make sure we don't try to get the value
        # of something if it doesn't look like a variable
        return_value = self.cli.default("something")
        self.assertIsNone(return_value)
        self.mock_builtin.get_variable.value.assert_not_called()

        # valid variable syntax, but unknown variable
        self.cli.stdout = io.StringIO()
        self.mock_builtin.get_variable_value.side_effect = Exception("not a variable")
        return_value = self.cli.default("${bogus}")
        self.assertEqual(self.cli.stdout.getvalue(), "unknown variable '${bogus}'\n")

    @mock.patch("pdb.Pdb")
    def test_pdb(self, mock_pdb):
        """Verify that the 'pdb' command starts pdb"""
        self.cli.do_pdb("")
        mock_pdb.assert_has_calls(
            [mock.call(stdout=self.cli.stdout), mock.call().set_trace()]
        )

    def test_continue(self):
        """Verify that the 'continue' debugger command returns a truthy value

        The truthy value is what triggers the command processor to exit so
        that robot can continue.
        """
        return_value = self.cli.do_continue("")
        self.assertTrue(return_value)

    def test_selenium(self):
        self.cli.selenium
        self.cli.builtin.get_library_instance.assert_called_with("SeleniumLibrary")

    @mock.patch.object(debugger.DebuggerCli, "selenium", mock.Mock())
    def test_locate_elements(self):
        """Test that the 'locate_elements' debugger command works

        This test sets up a mock that acts like selenium found several
        elements, and verifies that it calls the functions to highlight them
        """
        self.cli._highlight_element = mock.Mock()
        self.cli.selenium.get_webelements.return_value = ["Element1", "Element2"]
        return_value = self.cli.do_locate_elements("//whatever")
        self.assertIsNone(return_value)
        self.cli._highlight_element.assert_has_calls(
            [mock.call("Element1"), mock.call("Element2")]
        )
        self.assertEqual(self.cli.stdout.getvalue(), "Found 2 matches\n")

    @mock.patch.object(debugger.DebuggerCli, "selenium", mock.Mock())
    def test_locate_elements_exception_handling(self):
        """Verify that the 'locate_elements' debugger command handles exceptions"""

        self.cli._highlight_element = mock.Mock()
        self.cli.selenium.get_webelements.side_effect = InvalidSelectorException(
            "invalid xpath"
        )
        self.cli.stdout = io.StringIO()
        return_value = self.cli.do_locate_elements("//whatever")
        self.assertIsNone(return_value)
        self.cli._highlight_element.assert_not_called()
        self.assertEqual(self.cli.stdout.getvalue(), "invalid locator '//whatever'\n")

        # Even if get_webelement throws an exception, the keyword
        # should handle it gracefully
        self.cli.selenium.get_webelements.side_effect = Exception(
            "something unexpected"
        )
        self.cli.stdout = io.StringIO()
        return_value = self.cli.do_locate_elements("//whatever")
        self.assertIsNone(return_value)
        self.cli._highlight_element.assert_not_called()
        self.assertEqual(self.cli.stdout.getvalue(), "something unexpected\n")

    @mock.patch.object(debugger.DebuggerCli, "selenium", mock.Mock())
    def test_highlight_element_executes_javascript(self):
        mock_element = mock.Mock()
        mock_element.get_attribute.return_value = ""
        self.cli._highlight_element(mock_element)
        self.cli.selenium.driver.execute_script.assert_called()

    @mock.patch.object(debugger.DebuggerCli, "selenium", mock.Mock())
    def test_reset_elements(self):
        """Verify that the 'reset_elements' debugger command works"""
        self.cli._restore_element_style = mock.Mock()
        self.cli.selenium.get_webelements.return_value = ["Element1", "Element2"]
        return_value = self.cli.do_reset_elements("")
        self.assertIsNone(return_value)
        self.cli._restore_element_style.assert_has_calls(
            [mock.call("Element1"), mock.call("Element2")]
        )

    @mock.patch.object(debugger.DebuggerCli, "selenium", mock.Mock())
    def test_reset_elements_logging(self):
        self.cli.selenium.driver.execute_script.return_value = False
        self.cli._restore_element_style("dummy")
        self.cli.builtin.log.assert_called_with(
            "unable to restore style; original style not found", "DEBUG"
        )

    def test_shell_no_variables(self):
        """Verify that the shell command works"""
        with mock.patch.object(
            self.mock_builtin, "run_keyword_and_ignore_error", return_value=("PASS", 42)
        ):
            return_value = self.cli.do_shell("some keyword")
            self.assertIsNone(return_value)
            assert not self.mock_builtin.set_test_variable.called

    def test_shell_one_variable(self):
        """Verify that the shell command can set a single variable"""
        with mock.patch.object(
            self.mock_builtin, "run_keyword_and_ignore_error", return_value=("PASS", 42)
        ):

            return_value = self.cli.do_shell(
                "${value}  get variable value  ${whatever}"
            )
            self.assertIsNone(return_value)

            self.mock_builtin.run_keyword_and_ignore_error.assert_called_with(
                "get variable value", "${whatever}"
            )
            self.mock_builtin.set_test_variable.assert_called_with("${value}", 42)

    def test_shell_fail(self):
        """Verify that if a shell command fails, variables aren't set"""
        with mock.patch.object(
            self.mock_builtin,
            "run_keyword_and_ignore_error",
            return_value=("FAIL", (None,)),
        ):
            return_value = self.cli.do_shell(
                "${value1}  ${value2}   some keyword  ${whatever}"
            )
            self.assertIsNone(return_value)

            self.mock_builtin.set_test_variable.assert_not_called()

    def test_shell_two_variables(self):
        """Verify that the shell command can set more than one variable"""
        with mock.patch.object(
            self.mock_builtin,
            "run_keyword_and_ignore_error",
            return_value=("PASS", ("Inigo", "Montoya")),
        ):
            return_value = self.cli.do_shell(
                "${value1}  ${value2}   some keyword  ${whatever}"
            )
            self.assertIsNone(return_value)

            self.mock_builtin.set_test_variable.assert_has_calls(
                [mock.call("${value1}", "Inigo"), mock.call("${value2}", "Montoya")]
            )

    def test_shell_exception(self):
        with mock.patch.object(
            self.mock_builtin,
            "run_keyword_and_ignore_error",
            side_effect=Exception("Danger, Will Robinson!"),
        ):
            self.cli.do_shell("${value1}  ${value2}   some keyword  ${whatever}")
            self.assertEqual(
                self.cli.stdout.getvalue(),
                "error running keyword: Danger, Will Robinson!\n",
            )

    def test_step(self):
        """Test the 'step' debugger command"""
        # unlike most other commands, this one returns True
        # to allow the debugger to continue. We also need to verify
        # That the number of breakpoints after the step is the
        # same as before the step
        return_value = self.cli.do_step("")
        self.assertEqual(return_value, True)
        self.mock_listener.do_step.assert_called()

    def test_where(self):
        """Test the 'where' debugger command"""
        self.mock_listener.stack = [
            debugger.Suite(name="Suite", attrs={}),
            debugger.Suite(name="Another Suite", attrs={}),
            debugger.Testcase(name="Testcase", attrs={}),
            debugger.Keyword(name="A keyword", attrs={}),
        ]
        return_value = self.cli.do_where("")
        self.assertIsNone(return_value)

        # FIXME: add assertions to prove that the listener is
        # keeping track of the stack

    @mock.patch("cumulusci.tasks.robotframework.debugger.ui.CliTable")
    def test_vars(self, mock_clitbl):
        """Test the 'vars' debugger command"""
        self.mock_builtin.get_variables.return_value = {"one": 1, "two": 2}
        self.cli.do_vars("")
        mock_clitbl.assert_called_with([["Variable", "Value"], ["one", 1], ["two", 2]])
        mock_clitbl.return_value.echo.assert_called()

    def test_quit(self):
        with mock.patch("os.kill") as mock_kill:
            self.cli.do_quit("")
            mock_kill.assert_called_with(os.getpid(), signal.SIGTERM)


class TestInternalModels(unittest.TestCase):
    def test_testcase(self):
        testcase = debugger.Testcase(
            name="Test Case #1", attrs={"longname": "Root.Test Case #1"}
        )
        self.assertEqual(repr(testcase), "<Testcase: Test Case #1>")

        # robot passes in a longname, so make sure the property reflects it
        self.assertEqual(testcase.longname, "Root.Test Case #1")

    def test_keyword(self):
        keyword = debugger.Keyword(name="Keyword #1", attrs={"args": ["foo", "bar"]})
        self.assertEqual(repr(keyword), "<Keyword: Keyword #1  foo  bar>")
        # robot will NOT pass in a longname; make sure the property handles that case
        self.assertEqual(keyword.longname, keyword.name)

    def test_suite(self):
        suite = debugger.Suite(
            name="Suite #1", attrs={"source": "test.robot", "longname": "Root.Suite #1"}
        )
        self.assertEqual(repr(suite), "<Suite: Suite #1 (test.robot)>")
        # robot passes in a longname, so make sure the property reflects it
        self.assertEqual(suite.longname, "Root.Suite #1")

    def test_breakpoint_match(self):
        bp = debugger.Breakpoint(debugger.Keyword, "*::breakpoint")
        self.assertTrue(
            bp.match(context="Suite.Test Case::breakpoint"),
            "expected breakpoint to match, but it didn't",
        )
        self.assertFalse(
            bp.match(context="Suite.Test Case::some other keyword"),
            "didn't expect breakpoint to match, but it did",
        )



import sarge

from cumulusci.core.config import (
    BaseConfig,
    ConnectedAppOAuthConfig,
    ScratchOrgConfig,
    ServiceConfig,
)
from cumulusci.core.exceptions import (
    CumulusCIException,
    CumulusCIUsageError,
    OrgNotFound,
    ServiceNotConfigured,
    ServiceNotValid,
)
from cumulusci.core.sfdx import sfdx

DEFAULT_CONNECTED_APP_PORT = 7788
DEFAULT_CONNECTED_APP_NAME = "built-in"
DEFAULT_CONNECTED_APP = ConnectedAppOAuthConfig(
    {
        "client_id": "3MVG9i1HRpGLXp.or6OVlWVWyn8DXi9xueKNM4npq_AWh.yqswojK9sE5WY7f.biP0w7bNJIENfXc7JMDZGO1",
        "client_secret": None,
        "callback_url": f"http://localhost:{DEFAULT_CONNECTED_APP_PORT}/callback",
    }
)


class BaseProjectKeychain(BaseConfig):
    encrypted = False

    def __init__(self, project_config, key):
        super(BaseProjectKeychain, self).__init__()
        self.config = {
            "orgs": {},
            "app": None,
            "services": {},
        }
        self.project_config = project_config
        self._default_services = {}
        self.key = key
        self._validate_key()
        self._load_keychain()

    def _load_keychain(self):
        self._load_orgs()
        self._load_scratch_orgs()
        self._load_default_connected_app()
        self._load_services()
        self._load_default_services()

    def _validate_key(self):
        pass

    #######################################
    #               Orgs                  #
    #######################################

    def create_scratch_org(self, org_name, config_name, days=None, set_password=True):
        """Adds/Updates a scratch org config to the keychain from a named config"""
        scratch_config = getattr(self.project_config, f"orgs__scratch__{config_name}")
        if days is not None:
            # Allow override of scratch config's default days
            scratch_config["days"] = days
        else:
            # Use scratch config days or default of 1 day
            scratch_config.setdefault("days", 1)
        scratch_config["set_password"] = bool(set_password)
        scratch_config["scratch"] = True
        scratch_config.setdefault("namespaced", False)
        scratch_config["config_name"] = config_name
        scratch_config[
            "sfdx_alias"
        ] = f"{self.project_config.project__name}__{org_name}"
        org_config = ScratchOrgConfig(
            scratch_config, org_name, keychain=self, global_org=False
        )
        org_config.save()

    def set_org(self, org_config, global_org=False, save=True):
        if isinstance(org_config, ScratchOrgConfig):
            org_config.config["scratch"] = True
        self._set_org(org_config, global_org, save=save)

    def set_default_org(self, name):
        """set the default org for tasks and flows by name"""
        org = self.get_org(name)
        self.unset_default_org()
        org.config["default"] = True
        org.save()
        if org.created:
            sfdx(
                sarge.shell_format(
                    "force:config:set defaultusername={}", org.sfdx_alias
                )
            )

    def unset_default_org(self):
        """unset the default orgs for tasks"""
        for org in self.list_orgs():
            org_config = self.get_org(org)
            if org_config.default:
                del org_config.config["default"]
                org_config.save()
        sfdx("force:config:set defaultusername=")

    # This implementation of get_default_org, set_default_org, and unset_default_org
    # is currently kept for backwards compatibility, but EncryptedFileProjectKeychain
    # now stores the default elsewhere, and EnvironmentProjectKeychain doesn't actually
    # persist across multiple invocations of cci, so we should consider getting rid of this.

    def get_default_org(self):
        """retrieve the name and configuration of the default org"""
        for org in self.list_orgs():
            org_config = self.get_org(org)
            if org_config.default:
                return org, org_config
        return None, None

    def get_org(self, name: str):
        """retrieve an org configuration by name key"""
        org = self._get_org(name)
        if org.keychain:
            assert org.keychain is self
        else:
            org.keychain = self
        return org

    def list_orgs(self):
        """list the orgs configured in the keychain"""
        orgs = list(self.orgs.keys())
        orgs.sort()
        return orgs

    def remove_org(self, name, global_org=None):
        if name in self.orgs.keys():
            self._remove_org(name, global_org)
        self.cleanup_org_cache_dirs()

    def _load_orgs(self):
        pass

    def _load_scratch_orgs(self):
        """Creates all scratch org configs for the project in the keychain if
        a keychain org doesn't already exist"""
        current_orgs = self.list_orgs()
        if not self.project_config.orgs__scratch:
            return
        for config_name in self.project_config.orgs__scratch.keys():
            if config_name in current_orgs:
                # Don't overwrite an existing keychain org
                continue
            self.create_scratch_org(config_name, config_name)

    def _set_org(self, org_config, global_org, save=True):
        self.orgs[org_config.name] = org_config

    def _get_org(self, name):
        if name not in self.orgs:
            raise OrgNotFound(f"Org named {name} was not found in keychain")
        return self.orgs.get(name)

    def _remove_org(self, name, global_org):
        del self.orgs[name]
        self._load_orgs()

    def cleanup_org_cache_dirs(self):
        pass  # pragma: no cover

    #######################################
    #              Services               #
    #######################################

    def set_default_service(
        self, service_type: str, alias: str, project: bool = False, save: bool = True
    ) -> None:
        """Public API for setting a default service e.g. `cci service default`

        @param service_type: the type of service
        @param alias: the name of the service
        @param project: Should this be a project default
        @param save: save the defaults so they are loaded on subsequent executions
        @raises ServiceNotConfigured if service_type or alias are invalid
        """
        self._validate_service_type_and_alias(service_type, alias)
        self._default_services[service_type] = alias

    def get_default_service_name(self, service_type: str):
        """Returns the name of the default service for the given type
        or None if no default is currently set for the given type."""
        try:
            return self._default_services[service_type]
        except KeyError:
            return None

    def set_service(
        self,
        service_type: str,
        alias: str,
        service_config: ServiceConfig,
        save: bool = True,
        config_encrypted: bool = False,
    ):
        """Store a ServiceConfig in the keychain
        @service_type - type of service
        @alias - name that the service will be stored under
        @service_config - dict of service attributes
        @save - If true, indicates that the service
        should be saved in some manner (subclasses implement).
        @config_encrypted - Indicates whether or not the config
        is already encrypted (as can be the case when reading
        services back from an encrypted file).
        """
        # we only validate attributes when they aren't encrypted
        # if we are setting a service from an encrypted file then it has already been validated
        if not config_encrypted:
            self._validate_service(service_type, alias, service_config)
        self._set_service(
            service_type,
            alias,
            service_config,
            save=save,
            config_encrypted=config_encrypted,
        )

    def get_service(self, service_type, alias=None):
        """Retrieve a stored ServiceConfig from the keychain.
        If no alias is specified then the default service for
        the given type is returned.

        @param service_type: the service to retrieve e.g. 'github'
        @param alias: the alias of the service
        @returns: ServiceConfig for the requested service
        """
        if not self.project_config.services:
            raise ServiceNotValid(
                "Expecting services to be loaded, but none were found."
            )
        elif service_type not in self.project_config.services:
            raise ServiceNotConfigured(
                f"Service type is not configured: {service_type}"
            )

        if service_type not in self.services:
            self._raise_service_not_configured(service_type)

        if not alias:
            alias = self.get_default_service_name(service_type)
            if not alias:
                raise CumulusCIException(
                    f"No default service currently set for service type: {service_type}"
                )
        service = self._get_service(service_type, alias)

        # transparent migration of github API tokens to new key
        if service_type == "github" and service.password and not service.token:
            service.config["token"] = service.password

        return service

    def list_services(self):
        """list the services configured in the keychain"""
        service_types = list(self.services.keys())
        service_types.sort()

        services = {}
        for s_type in service_types:
            if s_type not in services:
                services[s_type] = []
            names = list(self.services[s_type].keys())
            names.sort()
            for name in names:
                services[s_type].append(name)
        return services

    def rename_service(
        self, service_type: str, current_alias: str, new_alias: str
    ) -> None:
        """Public API for renaming a service

        @param service_type type of service being renamed
        @param current_alias the current alias of the service
        @param new_alias the new alias for the service
        @throws: ServiceNotValid if no services of the given type are configured,
        or if no service of the given type has the current_alias
        """
        if (
            service_type == "connected_app"
            and current_alias == DEFAULT_CONNECTED_APP_NAME
        ):
            raise CumulusCIException(
                "You cannot rename the connected app service that is provided by CumulusCI."
            )

        self._validate_service_type_and_alias(service_type, current_alias)
        if new_alias in self.services[service_type]:
            raise CumulusCIUsageError(
                f"A service of type {service_type} already exists with name: {new_alias}"
            )

        self.services[service_type][new_alias] = self.services[service_type].pop(
            current_alias
        )

        if self._default_services.get(service_type) == current_alias:
            self._default_services[service_type] = new_alias

    def remove_service(self, service_type: str, alias: str):
        """Removes the given service from the keychain. If the service
        is the default service, and there is only one other service
        of the same type, that service is set as the new default.

        @param service_type type of the service
        @param alias the name of the service
        @raises ServiceNotConfigured if the service_type or alias are invalid
        """
        if service_type == "connected_app" and alias == DEFAULT_CONNECTED_APP_NAME:
            raise CumulusCIException(
                f"Unable to remove connected app service: {DEFAULT_CONNECTED_APP_NAME}. "
                "This connected app is provided by CumulusCI and cannot be removed."
            )

        self._validate_service_type_and_alias(service_type, alias)
        # remove the loaded service from the keychain
        del self.services[service_type][alias]

        # if set, remove the service as the default
        if alias == self._default_services[service_type]:
            del self._default_services[service_type]
            if len(self.services[service_type].keys()) == 1:
                alias = self.list_services()[service_type][0]
                self.set_default_service(service_type, alias, project=False)

    def _load_services(self):
        pass

    def _load_default_services(self):
        self._default_services["connected_app"] = DEFAULT_CONNECTED_APP_NAME

    def _load_default_connected_app(self):
        """Load the default connected app as a first class service on the keychain."""
        if "connected_app" not in self.config["services"]:
            self.config["services"]["connected_app"] = {}
        self.config["services"]["connected_app"][
            DEFAULT_CONNECTED_APP_NAME
        ] = DEFAULT_CONNECTED_APP

    def _set_service(
        self, service_type, alias, service_config, save=True, config_encrypted=False
    ):
        # The first service of a given service_type automatically becomes the default
        if service_type not in self.services:
            self.services[service_type] = {}
            self._default_services[service_type] = alias

        self.services[service_type][alias] = service_config

    def _get_service(self, service_type, alias):
        try:
            return self.services[service_type][alias]
        except KeyError:
            raise ServiceNotConfigured(
                f"No service of type {service_type} configured with name: {alias}"
            )

    def _validate_service(self, service_type, alias, config):
        if (
            not self.project_config.services
            or service_type not in self.project_config.services
        ):
            raise ServiceNotValid(
                f"Service named {alias} is not valid for this project"
            )

        self._validate_service_alias(service_type, alias)
        self._validate_service_attributes(service_type, config)

    def _validate_service_attributes(self, service_type, service_config):
        """
        Validate that all of the required attributes for a
        given service are present.
        """
        missing_required = []
        attr_key = f"services__{service_type}__attributes"
        for atr, config in list(getattr(self.project_config, attr_key).items()):
            if config.get("required") and not getattr(service_config, atr):
                missing_required.append(atr)

        if missing_required:
            if service_type == "github" and missing_required == ["token"]:
                service_config.token = service_config.password
                return

            raise ServiceNotValid(
                f"Missing required attribute(s) for {service_type} service: {missing_required}"
            )

    def _validate_service_alias(self, service_type, alias):
        if alias == service_type:
            raise ServiceNotValid(
                "Service name cannot be the same as the service type."
            )
        if service_type == "connected_app" and alias == DEFAULT_CONNECTED_APP_NAME:
            raise ServiceNotValid(
                f"You cannot use the name {DEFAULT_CONNECTED_APP_NAME} for a connected app service. Please select a different name."
            )

    def _validate_service_type_and_alias(self, service_type, alias):
        """Raises ServiceNotConfigured exception if the service_type
        or alias are not valid."""
        if service_type not in self.services:
            raise ServiceNotConfigured(
                f"No services of type {service_type} are currently configured"
            )
        elif alias not in self.services[service_type]:
            raise ServiceNotConfigured(
                f"No service of type {service_type} configured with the name: {alias}"
            )

    def _raise_service_not_configured(self, service_type):
        service_types = ", ".join(list(self.services))
        raise ServiceNotConfigured(
            f"Service type {service_type} is not configured for this project. Configured services are: {service_types}"
        )

    @property
    def cache_dir(self):
        "Helper function to get the cache_dir from the project_config"
        return self.project_config.cache_dir  # pragma: no cover



import abc

from robot.libraries.BuiltIn import BuiltIn
from selenium.common.exceptions import TimeoutException

from cumulusci.utils.classutils import get_all_subclasses

"""
FIXME: Something that still needs to be resolved is what to do about
name clashes. What if a project imports two different libraries and
each library implements a handler for the same tag and/or a handler
for which a built-in handler already exists. How do we decide what
takes precedence. Is this even something we have to worry about?

So, the idea is to find an element, then do something like
BaseFormHandler.get_handler(element, locator).set("foo")

"""


def get_form_handler(element, locator):
    """Return an instance of handler for the given element

    This will search all subclasses of BaseFormHandler, looking
    for a class that supports the tag_name of the given element
    """
    tag = element.tag_name
    for subclass in get_all_subclasses(BaseFormHandler):
        if tag in subclass.tags:
            # first one wins, but is that the right algorithm? Do
            # we want to give precedence to external handlers
            # over our built-in ones?
            return subclass(element, locator)
    return None


class BaseFormHandler(abc.ABC):
    """Base class for all form handlers

    The goal is to have each handler implement the methods 'get',
    'set', 'focus', and 'clear'. However, at present we're only using
    'set' externally. 'focus' is the one method that will probably
    have the same implementation for every handler so it's defined in
    the base class. The others are abstract methods which each handler
    should define.

    """

    def __init__(self, element, locator):
        # concrete element to be handled
        self.element = element

        # the locator used to get the element, which we use for error
        # handling.
        self.locator = locator

    @property
    def selenium(self):
        return BuiltIn().get_library_instance("SeleniumLibrary")

    @property
    def input_element(self):
        """Returns the actual <input> or <textarea> element inside the element"""
        elements = self.element.find_elements_by_xpath(
            ".//*[self::input or self::textarea]"
        )
        return elements[0] if elements else None

    @abc.abstractmethod
    def set(self, value):
        pass

    @abc.abstractmethod
    def get(self):
        pass

    @abc.abstractmethod
    def clear(self):
        pass

    def focus(self):
        """Set focus to the element

        In addition to merely setting the focus via selenium, we click
        on the element in case there are functions tied to that event.
        """
        self.selenium.set_focus_to_element(self.element)
        self.element.click()


class HTMLInputHandler(BaseFormHandler):
    """An input handler for non-lightning input and textarea form fields

    This is the fallback handler for when we can't find a lightning component.
    """

    tags = ["input", "textarea"]

    def set(self, value):
        self.element.send_keys(value)

    def get(self, value):
        # not currently being used
        pass  # pragma: no cover

    def clear(self):
        # Salesforce.py has a crazy amount of code for clearing
        # a field; for now these handlers seem to be working without it.
        self.selenium.driver.execute_script("arguments[0].value = '';", self.element)


class LightningComboboxHandler(BaseFormHandler):
    """An input handler for comboboxes"""

    tags = ["lightning-combobox"]

    @property
    def input_element(self):
        """Returns the base form element (input or button) that the combobox is based on"""
        elements = self.element.find_elements_by_xpath(
            ".//*[contains(@class, 'slds-combobox__input')]"
        )
        return elements[0] if elements else None

    def set(self, value):
        value_locator = f'//lightning-base-combobox-item[.="{value}"]'
        wait = 5
        try:
            # at this point, self.input_element is None
            self.input_element.click()
            self.selenium.wait_until_element_is_visible(value_locator, wait)
            self.selenium.click_element(value_locator)
        except Exception:
            raise TimeoutException(
                f"Dropdown value '{value}' for '{self.locator}' not found after {wait} seconds"
            )

    def get(self, value):
        # not currently being used
        pass  # pragma: no cover

    def clear(self):
        # not currently being used
        pass  # pragma: no cover


class LightningInputHandler(BaseFormHandler):
    """An input handler for components that can be treated as an input or textarea"""

    tags = ["lightning-input", "lightning-textarea", "lightning-datepicker"]

    def set(self, value):
        self.focus()
        if self.input_element.get_attribute("type") == "checkbox":
            # lightning-input elements are used for checkboxes
            # as well as free text input.
            checked = self.element.get_attribute("checked")
            if (checked and value != "checked") or (not checked and value == "checked"):
                self.input_element.send_keys(" ")
        else:
            self.clear()
            self.input_element.send_keys(value)

    def get(self, value):
        # not currently being used
        pass  # pragma: no cover

    def clear(self):
        # Salesforce.py has a crazy amount of code for clearing
        # a field; for now these handlers seem to be working without it.
        self.selenium.driver.execute_script("arguments[0].value = '';", self.element)


class LightningLookupHandler(BaseFormHandler):
    tags = ["lightning-lookup", "lightning-grouped-combobox"]

    def set(self, value):
        wait = 5
        # I wonder if I should/could anchor this to the element
        # instead of searching the whole document?
        value_locator = f'//div[@role="listbox"]//*[.="{value}"]'
        self.element.click()
        self.input_element.send_keys(value)
        try:
            self.selenium.wait_until_element_is_visible(value_locator, wait)
            self.selenium.click_element(value_locator)
        except Exception:
            # *sigh* this still fails randomly even though
            # I can see the dadgum item in the dropdown.
            raise TimeoutException(
                f"Lookup value '{value}' for '{self.locator}' not found after {wait} seconds"
            )

    def get(self, value):
        # not currently being used
        pass  # pragma: no cover

    def clear(self):
        # this is not currently being used
        pass  # pragma: no cover



import queue
import random
import shutil
import time
import typing as T
from collections import defaultdict
from multiprocessing import Lock
from pathlib import Path

import cumulusci.core.exceptions as exc
from cumulusci.core.config import OrgConfig
from cumulusci.tasks.bulkdata.generate_from_yaml import GenerateDataFromYaml
from cumulusci.tasks.bulkdata.load import LoadData
from cumulusci.utils.parallel.task_worker_queues.parallel_worker_queue import (
    WorkerQueue,
    WorkerQueueConfig,
)

from .snowfakery_run_until import PortionGenerator
from .snowfakery_working_directory import SnowfakeryWorkingDirectory
from .subtask_configurator import SubtaskConfigurator

# number of portions we will allow to be on-disk waiting to be loaded
# higher numbers use more disk space.
LOAD_QUEUE_SIZE = 15
# more loader workers than generators because they spend so much time
# waiting for responses. 4:1 is an experimentally derived ratio
WORKER_TO_LOADER_RATIO = 4


class SnowfakeryChannelManager:
    """The channels describe the 'shape' of the communication

    The normal case is a single, parallelized, bulk channel,
    multi-threaded on client and server, using a single user
    account.

    Using .load.yml you can add more channels, utilizing
    more user accounts which can speed up throughput in
    a few cases."""

    def __init__(
        self,
        subtask_configurator,
        *,
        project_config,
        logger,
    ):
        # Look at the docstring on get_results_report to understand
        # what this queue is for.
        #
        # Be careful to use a Queue class appropriate to
        # the spawn type (thread, process) you're using.
        #
        # Snowfakery runs its loader in threads, so queue.Queue()
        # works.
        #
        # multiprocessing.Manager().Queue() also seems to work,
        # and work across processes, (PR #3080) but it's dramatically
        # slower. See attachment to PR #3076
        self.results_reporter = queue.Queue()
        self.channels = []
        self.project_config = project_config
        self.logger = logger
        self.subtask_configurator = subtask_configurator
        self.start_time = time.time()

    def add_channel(
        self,
        org_config: OrgConfig,
        num_generator_workers: int,
        num_loader_workers: T.Optional[int],
        working_directory: Path,
        recipe_options: dict,
    ):
        if not num_loader_workers:
            num_loader_workers = num_generator_workers * WORKER_TO_LOADER_RATIO

        self.channels.append(
            Channel(
                project_config=self.project_config,
                org_config=org_config,
                num_generator_workers=num_generator_workers,
                num_loader_workers=num_loader_workers,
                working_directory=working_directory,
                subtask_configurator=self.subtask_configurator,
                logger=self.logger,
                results_reporter=self.results_reporter,
                recipe_options=recipe_options,
            )
        )

    def tick(
        self,
        upload_status,
        template_path: Path,
        tempdir: Path,
        portions: PortionGenerator,
        get_upload_status: T.Callable,
    ):
        """Called every few seconds, to make new data generators if needed."""
        # advance each channel's queues
        for channel in self.channels:
            channel.tick()

        # populate channels that have space
        self.assign_work_to_channels(
            upload_status,
            template_path,
            tempdir,
            portions,
            get_upload_status,
        )

    def assign_work_to_channels(
        self,
        upload_status,
        template_path: Path,
        tempdir: Path,
        portions: PortionGenerator,
        get_upload_status: T.Callable,
    ):
        def channel_with_free_space():
            return [channel for channel in self.channels if not channel.full]

        new_workers = [True]  # initial value to get into the while loop
        channels = channel_with_free_space()
        while (
            channels
            and any(new_workers)
            and not portions.done(upload_status.total_sets_working_on_or_uploaded)
        ):
            # randomize channels so the first one doesn't always grab all slots
            random.shuffle(channels)
            new_workers = [
                channel.make_new_worker(
                    template_path, tempdir, portions, get_upload_status
                )
                for channel in channels
            ]
            channels = channel_with_free_space()

    def get_upload_status(self, batch_size, sets_finished_while_generating_template):
        """Combine information from the different data sources into a single "report".

        Useful for debugging but also for making decisions about what to do next."""
        summed_statuses = defaultdict(int)
        for channel in self.channels:
            for key, value in channel.get_upload_status_for_channel().items():
                summed_statuses[key] += value

        summed_statuses["sets_finished"] += sets_finished_while_generating_template
        return UploadStatus(
            target_count=self.subtask_configurator.run_until.gap,
            elapsed_seconds=int(self.elapsed_seconds()),
            batch_size=batch_size,
            channels=len(self.channels),
            **summed_statuses,
        )

    def elapsed_seconds(self):
        return time.time() - self.start_time

    def failure_descriptions(self) -> T.List[str]:
        """Log failures from sub-processes to main process"""
        ret = []
        for channel in self.channels:
            ret.extend(channel.failure_descriptions())
        return ret

    def check_finished(self) -> bool:
        for channel in self.channels:
            channel.tick()
        return all([channel.check_finished() for channel in self.channels])

    def get_results_report(self, block=False):
        """
        This is a realtime reporting channel which could, in theory, be updated
        before sub-tasks finish. Currently no sub-tasks are coded to do that.

        The logical next step is to allow LoadData to monitor steps one by
        one or even batches one by one.

        Note that until we implement that, we are paying the complexity
        cost of a real-time channel but not getting the benefits of it."""
        return self.results_reporter.get(block=block)


class Channel:
    def __init__(
        self,
        *,
        project_config,
        org_config,
        num_generator_workers: int,
        num_loader_workers: int,
        working_directory: Path,
        subtask_configurator: SubtaskConfigurator,
        logger,
        recipe_options=None,
        results_reporter=None,
    ):
        self.project_config = project_config
        self.org_config = org_config
        self.num_generator_workers = num_generator_workers
        self.num_loader_workers = num_loader_workers
        self.working_directory = working_directory
        self.subtask_configurator = subtask_configurator
        self.run_until = subtask_configurator.run_until
        self.logger = logger
        self.results_reporter = results_reporter
        self.filesystem_lock = Lock()
        self.job_counter = 0
        recipe_options = recipe_options or {}
        self._configure_queues(recipe_options)

    def _configure_queues(self, recipe_options):
        """Configure two ParallelWorkerQueues for datagen and dataload"""
        try:
            # in the future, the connected_app should come from the org
            connected_app = self.project_config.keychain.get_service("connected_app")
        except exc.ServiceNotConfigured:  # pragma: no cover
            # to discuss...when can this happen? What are the consequences?
            connected_app = None

        def data_generator_opts_callback(*args, **kwargs):
            return self.subtask_configurator.data_generator_opts(
                *args, recipe_options=recipe_options, **kwargs
            )

        data_gen_q_config = WorkerQueueConfig(
            project_config=self.project_config,
            org_config=self.org_config,
            connected_app=connected_app,
            redirect_logging=True,
            # processes are better for compute-heavy tasks (in Python)
            spawn_class=WorkerQueue.Process,
            parent_dir=self.working_directory,
            name="data_gen",
            task_class=GenerateDataFromYaml,
            make_task_options=data_generator_opts_callback,
            queue_size=0,
            num_workers=self.num_generator_workers,
        )
        # datagen queues do not get a result reporter because
        # a) we are less curious about how many records have
        #    been generated than we aare about how many are loaded
        # b) finding a queue type which is not prone to race conditions
        #    or perf slowdowns with "Process" task types (sub-processes)
        # is a challenge.
        self.data_gen_q = WorkerQueue(data_gen_q_config, self.filesystem_lock)

        load_data_q_config = WorkerQueueConfig(
            project_config=self.project_config,
            org_config=self.org_config,
            connected_app=connected_app,
            redirect_logging=True,
            spawn_class=WorkerQueue.Thread,
            parent_dir=self.working_directory,
            name="data_load",
            task_class=LoadData,
            make_task_options=self.subtask_configurator.data_loader_opts,
            queue_size=LOAD_QUEUE_SIZE,
            num_workers=self.num_loader_workers,
            rename_directory=self.data_loader_new_directory_name,
        )
        self.load_data_q = WorkerQueue(
            load_data_q_config, self.filesystem_lock, self.results_reporter
        )

        self.data_gen_q.feeds_data_to(self.load_data_q)
        return self.data_gen_q, self.load_data_q

    def data_loader_new_directory_name(self, working_directory):
        return data_loader_new_directory_name(working_directory, self.run_until)

    def tick(
        self,
    ):
        """Called every few seconds, to make new data generators if needed."""
        self.data_gen_q.tick()

    @property
    def full(self):
        return self.data_gen_q.full

    def make_new_worker(
        self,
        template_path: Path,
        tempdir: Path,
        portions: PortionGenerator,
        get_upload_status: T.Callable,
    ):
        if (
            self.data_gen_q.num_free_workers and self.data_gen_q.full
        ):  # pragma: no cover
            # TODO: investigate the consequences of taking this branch out
            self.logger.info("Waiting before datagen (load queue is full)")
        else:
            upload_status = get_upload_status(
                portions.next_batch_size,
            )
            self.job_counter += 1
            batch_size = portions.next_batch(
                upload_status.total_sets_working_on_or_uploaded
            )
            if not batch_size:
                self.logger.info("All scheduled portions generated and being uploaded")
                return

            job_dir = self.generator_data_dir(
                self.job_counter, template_path, batch_size, tempdir
            )
            self.data_gen_q.push(job_dir)
            return job_dir

    def generator_data_dir(self, idx, template_path, batch_size, parent_dir):
        """Create a new generator directory with a name based on index and batch_size"""
        assert batch_size > 0
        data_dir = parent_dir / (str(idx) + "_" + str(batch_size))
        shutil.copytree(template_path, data_dir)
        return data_dir

    def get_upload_status_for_channel(self):
        with self.filesystem_lock:

            def set_count_from_names(names):
                return sum(int(name.split("_")[1]) for name in names)

            return {
                "sets_queued_to_be_generated": set_count_from_names(
                    self.data_gen_q.queued_jobs
                ),
                "sets_being_generated": set_count_from_names(
                    self.data_gen_q.inprogress_jobs
                ),
                "sets_queued_for_loading": set_count_from_names(
                    self.load_data_q.queued_jobs
                ),
                # note that these may count as already imported in the org
                "sets_being_loaded": set_count_from_names(
                    self.load_data_q.inprogress_jobs
                ),
                "max_num_loader_workers": self.num_loader_workers,
                "max_num_generator_workers": self.num_generator_workers,
                # todo: use row-level result from org load for better accuracy
                "sets_finished": set_count_from_names(self.load_data_q.outbox_jobs),
                "sets_failed": len(self.load_data_q.failed_jobs),
                # TODO: are these two redundant?
                "inprogress_generator_jobs": len(self.data_gen_q.inprogress_jobs),
                "inprogress_loader_jobs": len(self.load_data_q.inprogress_jobs),
                "data_gen_free_workers": self.data_gen_q.num_free_workers,
            }

    def failure_descriptions(self) -> T.List[str]:
        """Log failures from sub-processes to main process"""
        failure_dirs = set(
            self.load_data_q.failed_job_dirs + self.data_gen_q.failed_job_dirs
        )

        def error_from_dir(failure_dir: Path) -> T.Optional[str]:
            f = Path(failure_dir) / "exception.txt"
            if not f.exists():  # pragma: no cover
                return
            return f.read_text(encoding="utf-8").strip().split("\n")[-1]

        errors = [error_from_dir(failure_dir) for failure_dir in failure_dirs]
        return [error for error in errors if error is not None]

    def check_finished(self) -> bool:
        self.data_gen_q.tick()
        with self.filesystem_lock:
            still_running = (
                len(
                    self.data_gen_q.workers
                    + self.data_gen_q.queued_job_dirs
                    + self.data_gen_q.inprogress_jobs
                    + self.load_data_q.workers
                    + self.load_data_q.inprogress_jobs
                    + self.load_data_q.queued_job_dirs
                )
                > 0
            )
        return not still_running


# TODO: This function is actually based on the number generated,
#       because it is called before the load.
#       If there are row errors, it will drift out of correctness
#       Code needs to be updated to rename again after load.
#       Or move away from using directory names for math altogether.
def data_loader_new_directory_name(working_dir: Path, run_until):
    """Change the directory name to reflect the true number of sets created."""

    if not run_until.sobject_name:
        return working_dir

    wd = SnowfakeryWorkingDirectory(working_dir)
    count = wd.get_record_counts()[run_until.sobject_name]

    path, _ = str(working_dir).rsplit("_", 1)
    new_working_dir = Path(path + "_" + str(count))
    return new_working_dir


class UploadStatus(T.NamedTuple):
    """Single "report" of the current status of our processes."""

    batch_size: int
    sets_being_generated: int
    sets_queued_to_be_generated: int
    sets_being_loaded: int
    sets_queued_for_loading: int
    sets_finished: int
    target_count: int
    max_num_loader_workers: int
    max_num_generator_workers: int
    elapsed_seconds: int
    sets_failed: int
    inprogress_generator_jobs: int
    inprogress_loader_jobs: int
    data_gen_free_workers: int
    channels: int

    @property
    def total_in_flight(self):
        return (
            self.sets_queued_to_be_generated
            + self.sets_being_generated
            + self.sets_queued_for_loading
            + self.sets_being_loaded
        )

    @property
    def total_sets_working_on_or_uploaded(
        self,
    ):
        return self.total_in_flight + self.sets_finished

    def _display(self, detailed=False):
        most_important_stats = [
            "target_count",
            "total_sets_working_on_or_uploaded",
            "sets_finished",
            "sets_failed",
        ]
        if self.channels > 1:
            most_important_stats.append("channels")

        queue_stats = [
            "inprogress_generator_jobs",
            "inprogress_loader_jobs",
        ]

        def display_stats(keys):
            def format(a):
                return f"{a.replace('_', ' ').title()}: {getattr(self, a):,}"

            return (
                "\n"
                + "\n".join(
                    format(a)
                    for a in keys
                    if not a[0] == "_" and not callable(getattr(self, a))
                )
                + "\n"
            )

        rc = display_stats(most_important_stats)
        rc += "\n   ** Workers **\n"
        rc += display_stats(queue_stats)
        if detailed:
            rc += "\n   ** Internals **\n"
            rc += display_stats(
                set(dir(self)) - (set(most_important_stats) & set(queue_stats))
            )
        return rc



from cumulusci.tasks.bulkdata.delete import DeleteData
from cumulusci.tasks.bulkdata.extract import ExtractData
from cumulusci.tasks.bulkdata.generate_mapping import GenerateMapping
from cumulusci.tasks.bulkdata.load import LoadData
from cumulusci.tasks.bulkdata.generate_and_load_data import GenerateAndLoadData

# For backwards-compatibility
QueryData = ExtractData
__all__ = (
    "DeleteData",
    "ExtractData",
    "GenerateMapping",
    "LoadData",
    "GenerateAndLoadData",
)



import gzip
import json
import re
from itertools import chain
from pathlib import Path
from unittest.mock import patch

import pytest
import responses
import yaml
from sqlalchemy import create_engine

from cumulusci.salesforce_api.org_schema import (
    BufferedSession,
    get_org_schema,
    zip_database,
)
from cumulusci.salesforce_api.org_schema_models import Base
from cumulusci.tests.util import FakeUnreliableRequestHandler


class FakeSF:
    """A pretend version of Salesforce for composite testing"""

    sf_version = "99.0"

    base_url = "https://innovation-page-2420-dev-ed.cs50.my.salesforce.com/"
    headers = {}

    def describe(self):
        return {
            "encoding": "UTF-8",
            "maxBatchSize": 200,
            "sobjects": [
                {"createable": False, "deletable": True, "name": "Account"},
                {"createable": False, "deletable": True, "name": "Contact"},
            ],
        }


def makeFakeCompositeParallelSalesforce(responses):
    class FakeCompositeParallelSalesforce:
        def __init__(self, sf, *args, **kwargs):
            self.sf = sf

        def __enter__(self, *args, **kwargs):
            return self

        def __exit__(self, *args, **kwargs):
            pass

        def do_composite_requests(self, requests):
            return responses(), []

    return FakeCompositeParallelSalesforce


def uncached_responses(responses):
    """Pretend to load uncached responses from a VCR cassette"""

    def parse_composite_response(interaction: dict):
        response_body = interaction["response"]["body"]["string"]
        struct = json.loads(response_body)
        return struct["compositeResponse"]

    return chain.from_iterable(
        parse_composite_response(interaction)
        for interaction in responses["interactions"]
        if interaction["request"]["uri"].endswith("/composite")
    )


cached_responses = [
    {"body": None, "httpHeaders": {}, "httpStatusCode": 304, "referenceId": "noId"}
] * 4


def mock_return_uncached_responses(cassette_data):
    return patch(
        "cumulusci.salesforce_api.org_schema.CompositeParallelSalesforce",
        makeFakeCompositeParallelSalesforce(lambda: uncached_responses(cassette_data)),
    )


def mock_return_cached_responses():
    return patch(
        "cumulusci.salesforce_api.org_schema.CompositeParallelSalesforce",
        makeFakeCompositeParallelSalesforce(lambda: cached_responses.copy()),
    )


class TestDescribeOrg:
    def setup_class(self):
        cassette = (
            Path(__file__).parent / "cassettes/ManualEdit_test_describe_to_sql.yaml"
        )
        with open(cassette) as f:
            self.cassette_data = yaml.safe_load(f)

    def validate_schema_data(self, schema):
        assert len(list(schema.sobjects)) == 4
        assert schema["Account"].createable is True
        assert schema["Account"].fields["Id"].aggregatable is True
        assert schema["Account"].labelPlural == "Accounts"
        account_desc = schema["Account"]
        for name, value in account_data.items():
            if name != "fields":
                assert account_desc[name] == value, (
                    name,
                    value,
                    account_desc[name],
                )
        for field in account_data["fields"]:
            db_field = account_desc["fields"][field["name"]]
            for name, value in field.items():
                assert db_field[name] == value, (
                    name,
                    value,
                    db_field[name],
                )

    def test_describe_to_sql(self, fallback_org_config):
        # Step 1: Pretend to download data from server
        org_config = fallback_org_config()
        with mock_return_uncached_responses(self.cassette_data):
            with get_org_schema(FakeSF(), org_config) as schema:
                self.validate_schema_data(schema)

        # Step 2: Call the server again.
        #         This time it has nothing new to tell us so nothing
        # should be written to the local database except an updated
        # LastModifiedDate.
        with mock_return_cached_responses(), patch(
            "cumulusci.salesforce_api.org_schema.create_row"
        ) as create_row, get_org_schema(FakeSF(), org_config) as schema:
            self.validate_schema_data(schema)
            for call in create_row.mock_calls:
                assert call[1][1].__name__ == "FileMetadata"

    def test_errors(self, org_config):
        with mock_return_uncached_responses(self.cassette_data), get_org_schema(
            FakeSF(), org_config
        ) as schema:
            with pytest.raises(KeyError):
                schema["Foo"]

    def test_forced_recache(self, org_config):
        with mock_return_uncached_responses(self.cassette_data):
            with get_org_schema(FakeSF(), org_config) as schema:
                schema.session.execute("insert into sobjects (name) values ('Foo')")
                assert "Foo" in schema
                schema.session._real_commit__()
                dbpath = schema.engine.url.translate_connect_args()["database"]
                zip_database(Path(dbpath), schema.path)
            with get_org_schema(FakeSF(), org_config) as schema:
                assert "Foo" in schema
            with get_org_schema(FakeSF(), org_config, force_recache=True) as schema:
                assert "Foo" not in schema

    def test_dict_like(self, org_config):
        with mock_return_uncached_responses(self.cassette_data):
            with get_org_schema(FakeSF(), org_config) as schema:
                assert schema["Account"]
                assert "Account" in schema
                assert sorted(schema.keys())[0] == "Account"
                assert (
                    sorted(schema.values(), key=lambda x: x.name)[0].name == "Account"
                )
                a, b = sorted(schema.items())[0]
                assert a == "Account"
                assert "Account" in schema.keys()
                assert "<Schema" in repr(schema)

    def test_misuse(self, org_config):
        """What if the user keeps a reference to the schema"""
        with mock_return_uncached_responses(self.cassette_data):
            with get_org_schema(FakeSF(), org_config) as schema:
                pass
        with pytest.raises(IOError):
            schema.session.commit()

    def test_corrupted_schema(self, caplog, org_config):
        "What if the schema GZip is corrupted?"
        with mock_return_uncached_responses(self.cassette_data):
            with get_org_schema(FakeSF(), org_config) as schema:
                assert "Account" in schema
                path = schema.path
            assert not caplog.text
            with open(path, "w") as p:
                p.write("xxx")

            with get_org_schema(FakeSF(), org_config) as schema:
                assert "Account" in schema
            assert caplog.text

    def test_corrupted_schema__sqlite(self, caplog, org_config):
        "What if the schema inside the gzip is corrupted"
        with mock_return_uncached_responses(self.cassette_data):
            with get_org_schema(FakeSF(), org_config) as schema:
                assert "Account" in schema
                path = schema.path
            assert not caplog.text
            with open(path, "wb") as p:
                with gzip.GzipFile(fileobj=p, mode="w") as gzipped:
                    gzipped.write(b"xxx")

            with get_org_schema(FakeSF(), org_config) as schema:
                assert "Account" in schema
            assert caplog.text

    @responses.activate
    def test_http_level_errors(self, sf, org_config, global_describe):
        # This is a bit complex. We're trying to test what happens
        # when a composite request fails. That should trigger a retry
        # of each item in the request. So we have to provide responses
        # at the level of the original describe, the composite request and
        # the single requests.

        # use just a subset for test perf reasons
        responses.add("GET", f"{sf.base_url}sobjects", json=global_describe(50))

        sobject_describes = json.loads(
            self.cassette_data["interactions"][0]["response"]["body"]["string"]
        )
        composite_handler = FakeUnreliableRequestHandler(sobject_describes)
        responses.add_callback(
            method="POST",
            url=f"{sf.base_url}composite",
            callback=composite_handler.request_callback,
            content_type="application/json",
        )

        class SingleSobjFakeUnreliableRequestHandler(FakeUnreliableRequestHandler):
            def real_reliable_request_callback(self, request):
                sobject = request.url.split("/")[-2]
                fake_describe = sobject_describes["compositeResponse"][0]["body"].copy()
                fake_describe["name"] = sobject
                return fake_describe

        single_sobj_handler = SingleSobjFakeUnreliableRequestHandler()
        responses.add_callback(
            method="GET",
            url=re.compile(r"https://.*/sobjects/.*/describe"),
            callback=single_sobj_handler.request_callback,
            content_type="application/json",
        )

        with get_org_schema(sf, org_config) as schema:
            assert "Account" in schema
            assert schema["Account"].labelPlural

        # check that the single_sobj_handler was called because of
        # the timeout. If you comment out the branch line
        # that raises the exception then this assertion will fail.
        assert single_sobj_handler.counter > 0

    @responses.activate
    def test_http_level_errors_after_retries(self, sf, org_config, global_describe):
        # This is a bit complex. We're trying to test what happens
        # when a composite request fails. That should trigger a retry
        # of each item in the request. So we have to provide responses
        # at the level of the original describe, the composite request and
        # the single requests.

        # use just a subset for test perf reasons
        responses.add("GET", f"{sf.base_url}sobjects", json=global_describe(50))

        sobject_describes = json.loads(
            self.cassette_data["interactions"][0]["response"]["body"]["string"]
        )
        composite_handler = FakeUnreliableRequestHandler(sobject_describes)
        responses.add_callback(
            method="POST",
            url=f"{sf.base_url}composite",
            callback=composite_handler.request_callback,
            content_type="application/json",
        )

        def throw_exception(*args, **kwargs):
            raise AssertionError()

        responses.add_callback(
            method="GET",
            url=re.compile(r"https://.*/sobjects/.*/describe"),
            callback=throw_exception,
            content_type="application/json",
        )

        with pytest.raises(AssertionError):
            with get_org_schema(sf, org_config):
                pass


@pytest.mark.needs_org()  # too hard to make these VCR-compatible due to data volume
@pytest.mark.slow()
class TestOrgSchemaIntegration:
    def validate_real_schema_data(self, schema):
        assert len(list(schema.sobjects)) > 800
        assert schema["Account"].createable is True
        assert schema["Account"].fields["Id"].aggregatable is True
        assert schema["Account"].labelPlural == "Accounts"
        account_desc = schema["Account"]
        assert account_desc.deletable is True
        assert account_desc.undeletable is True
        assert account_desc.keyPrefix == "001"

    def test_cache_schema(self, sf, org_config):
        with get_org_schema(sf, org_config, force_recache=True) as schema:
            self.validate_real_schema_data(schema)
            assert not schema.from_cache
        with get_org_schema(sf, org_config) as schema:
            self.validate_real_schema_data(schema)
            assert schema.from_cache


class TestBufferedSession:
    def test_buffer_empties(self):
        engine = create_engine("sqlite:///")
        Base.metadata.bind = engine
        Base.metadata.create_all()
        bs = BufferedSession(engine, Base.metadata, 5)
        with patch.object(bs, "flush") as flush:
            for i in range(0, 3):
                bs.write_single_row("sobjects", {"name": str(i)})
            assert not flush.mock_calls
            for i in range(0, 3):
                bs.write_single_row("sobjects", {"name": str(i)})

        assert flush.mock_calls


account_data = {
    "actionOverrides": (),
    "activateable": False,
    "compactLayoutable": True,
    "createable": True,
    "custom": False,
    "customSetting": False,
    "deepCloneable": False,
    "defaultImplementation": None,
    "deletable": True,
    "deprecatedAndHidden": False,
    "extendedBy": None,
    "extendsInterfaces": None,
    "feedEnabled": True,
    "fields": [
        {
            "aggregatable": True,
            "aiPredictionField": False,
            "autoNumber": False,
            "byteLength": 18,
            "calculated": False,
            "calculatedFormula": None,
            "cascadeDelete": False,
            "caseSensitive": False,
            "compoundFieldName": None,
            "controllerName": None,
            "createable": False,
            "custom": False,
            "defaultValue": None,
            "defaultValueFormula": None,
            "defaultedOnCreate": True,
            "dependentPicklist": False,
            "deprecatedAndHidden": False,
            "digits": 0,
            "displayLocationInDecimal": False,
            "encrypted": False,
            "externalId": False,
            "extraTypeInfo": None,
            "filterable": True,
            "filteredLookupInfo": None,
            "formulaTreatNullNumberAsZero": False,
            "groupable": True,
            "highScaleNumber": False,
            "htmlFormatted": False,
            "idLookup": True,
            "inlineHelpText": None,
            "label": "Account ID",
            "length": 18,
            "mask": None,
            "maskType": None,
            "name": "Id",
            "nameField": False,
            "namePointing": False,
            "nillable": False,
            "permissionable": False,
            "picklistValues": [],
            "polymorphicForeignKey": False,
            "precision": 0,
            "queryByDistance": False,
            "referenceTargetField": None,
            "referenceTo": [],
            "relationshipName": None,
            "relationshipOrder": None,
            "restrictedDelete": False,
            "restrictedPicklist": False,
            "scale": 0,
            "searchPrefilterable": False,
            "soapType": "tns:ID",
            "sortable": True,
            "type": "id",
            "unique": False,
            "updateable": False,
            "writeRequiresMasterRead": False,
        }
    ],
    "hasSubtypes": False,
    "implementedBy": None,
    "implementsInterfaces": None,
    "isInterface": False,
    "isSubtype": False,
    "keyPrefix": "001",
    "label": "Account",
    "labelPlural": "Accounts",
    "layoutable": True,
    "listviewable": None,
    "lookupLayoutable": None,
    "mergeable": True,
    "mruEnabled": True,
    "name": "Account",
    "namedLayoutInfos": (),
    "networkScopeFieldName": None,
    "queryable": True,
    "replicateable": True,
    "retrieveable": True,
    "searchLayoutable": True,
    "searchable": True,
    "sobjectDescribeOption": "FULL",
    "triggerable": True,
    "undeletable": True,
    "updateable": True,
    "urls": {"a": "b"},
}


## These are helper functions for managing the sample data used above.
## The functions aren't called above, but test maintainers could use
## them to update sample data.


def reduce_data(filename):
    """A function for reducing Casettes of Describes() to the bare minimum

    Note that the deep_describe infrastructure uses threads and may not
    always play nicely with vcr. It may take some experimentation and cutting
    and pasting to get a decent vcr cassette to use as the input of this
    function."""
    relevant_objs = ["Account", "Contact", "Opportunity", "Campaign"]
    data = yaml.safe_load(open(filename))

    for interaction in data["interactions"]:
        url = interaction["request"].get("uri")
        if url and url.endswith("composite"):
            reduce_composite(interaction)
        elif url:
            assert url.endswith("sobjects"), url
            reduce_describe(interaction, relevant_objs)

    yaml.safe_dump(data, open(filename + ".out.yaml", "w"))


def reduce_composite(interaction):
    results = json.loads(interaction["response"]["body"]["string"])["compositeResponse"]
    sobjs = [result["body"] for result in results]
    for sobj in sobjs:
        sobj["fields"] = sobj["fields"][0:1]
        for key in ["urls", "childRelationships", "recordTypeInfos", "supportedScopes"]:
            if sobj.get(key):
                del sobj[key]

    interaction["response"]["body"]["string"] = json.dumps(
        {"compositeResponse": results}
    )


def reduce_describe(interaction, relevant_objs):
    sobjs = json.loads(interaction["response"]["body"]["string"])["sobjects"]
    results = []
    for sobj in sobjs:
        if sobj["name"] not in relevant_objs:
            continue
        for key in ["urls", "childRelationships", "recordTypeInfos", "supportedScopes"]:
            if sobj.get(key):
                del sobj[key]
        results.append(sobj)

    interaction["response"]["body"]["string"] = json.dumps({"sobjects": results})
